#! /usr/bin/env python
# options.py

__all__ = [\
    "DiscOptions",
    "MenuOptions",
    "SlideOptions",
    "VideoOptions",
    "GroupOptions"]

import os
import wx

import libtovid
from libtovid.author import Group
from libtovid.gui.constants import *
from libtovid.gui.configs import TovidConfig
from libtovid.gui.util import _

class DiscOptions:
    """Options that apply to an entire disc"""
    type = ID_DISC
    format = 'dvd'
    tvsys = 'ntsc'
    title = "Untitled disc"

    def __init__(self, format = 'dvd', tvsys = 'ntsc'):
        self.format = format
        self.tvsys = tvsys

    def SetLayout(self, optionList):
        """Set disc layout hierarchy, given a list of VideoOptions,
        MenuOptions, and SlideOptions."""
        self.optionList = optionList

    def GetCommand(self):
        """Return the 'makexml' command for generating XML for this disc."""
        # Tricky part of this is know when to add -endgroup!
        amInAGroup = False
        videosInGroup = 0

        # Get global configuration (for output directory)
        curConfig = TovidConfig()

        strCommand = "makexml -quiet -overwrite -%s " % self.format

        for curItem in self.optionList:
            # Prefix -topmenu or -menu if necessary
            if curItem.type == ID_MENU:
                if curItem.isTopMenu:
                    strCommand += "-topmenu "
                else:
                    strCommand += "-menu "
            elif curItem.type == ID_GROUP:
                strCommand += "-group "
                amInAGroup = True
                videosInGroup = curItem.groupMemberCount

            if curItem.type != ID_GROUP:
                # Output path and encoded filename for all but groups
                strCommand += "\"%s/%s.mpg\" " % \
                    (curConfig.strOutputDirectory, curItem.outPrefix)
            if amInAGroup == True:
                #Need to work out whether need to endgroup
                if videosInGroup == 0:
                    # Cannot be in a group anymore
                    strCommand += "-endgroup "
                    amInAGroup = False
                else:
                    videosInGroup = videosInGroup - 1
                
        # Use output filename based on disc title
        self.outPrefix = self.title.replace(' ', '_')

        # Save output filename in global Config
        curConfig.strOutputXMLFile = "%s/%s" % (curConfig.strOutputDirectory,
            self.outPrefix)
        curConfig.curDiscFormat = self.format

        # Use output directory
        strCommand += "-out \"%s/%s\"" % \
                   (curConfig.strOutputDirectory, self.outPrefix)
        return strCommand

    def DiscNameOK(self, panel):
        """"Ensure there are no problems with the disc name"""
        # Get global configuration (for output directory)
        curConfig = TovidConfig()
        # Use output filename based on disc title
        self.outPrefix = self.title.replace(' ', '_')

        # Save output filename in global Config
        discName = "%s/%s.xml" % (curConfig.strOutputDirectory, self.outPrefix) 
        
        #If specified, check whether it exists
        if os.path.exists(discName) == True:
            #If exists, check whether it is a file or directory
            if os.path.isfile(discName) == True:
                msgDiscFileExistsDlg = wx.MessageDialog(panel, \
                    "A file needs to be created based on the disc name.\n" \
                    "But, this file (based on the disc name) already exists.\n"\
                    "The file is: %s\n\n" \
                    "Do you want to continue?\n"
                    "This will overwrite this existing file." % (discName),
                    "Existing Disc Name File",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
                response = msgDiscFileExistsDlg.ShowModal()
                msgDiscFileExistsDlg.Destroy()
                if response != wx.ID_YES:
                    return False
            else:   
                msgDiscFolderExistsDlg = wx.MessageDialog(panel, \
                    "A file needs to be created based on the disc name.\n" \
                    "But, a directory with the same name already exists.\n" \
                    "This directory is: %s\n\n" \
                    "Please either rename/remove the directory or " \
                    "change the disc name!" % (discName),
                    "Existing Directory",
                    wx.OK | wx.ICON_ERROR)
                msgDiscFolderExistsDlg.ShowModal()
                msgDiscFolderExistsDlg.Destroy()
                return False   

        #ICheck whether contains problematic characters
        if discName.find("'") > -1 or discName.find("\"") > -1 \
           or discName.find("$") > -1 or discName.find("&") > -1:

            #If so, give option of going back or trying anyway
            #NB, it is not just Python files that need fixing. 
            #Many lines throughout makemenu do aswell.
            msgDiscNameErrorDlg = wx.MessageDialog(panel, \
                "For technical reasons, currently apostrophes, double quotes\n"\
                "$ signs and & signs may cause problems when in the disc name.\n" \
                "The disc name is: %s\n\n" \
                "Do you want to return to rename the disc?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (self.title),
                "Problematic character in Disc Name",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgDiscNameErrorDlg.ShowModal()
            msgDiscNameErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        #If get this far, file is OK
        return True

    def OutputDirectoryExists(self, panel):
        """"Ensure there are no problems with the output directory name"""
        # Get global configuration (for output directory)
        curConfig = TovidConfig()
        outputDirectory = "%s" % (curConfig.strOutputDirectory)
 
        #If specified, check whether it exists
        if os.path.exists(outputDirectory) == True:
            #If exists, check whether it is a file or directory
            if os.path.isfile(outputDirectory) == True:
                msgOutputDirectoryIsAFileDlg = wx.MessageDialog(panel, \
                    "The output directory cannot be created as a file\n" \
                    "already exists in that location.\n"\
                    "The file is: %s\n\n" \
                    "Please change the output directory." % (outputDirectory),
                    "Output Directory Error",
                    wx.OK | wx.ICON_ERROR)
                msgOutputDirectoryIsAFileDlg.ShowModal()
                msgOutputDirectoryIsAFileDlg.Destroy()
                return False
        else:   
            msgOutputDirectoryDoesNotExistDlg = wx.MessageDialog(panel, \
                "The specified output directory does not exist.\n" \
                "The specified output directory is: %s\n\n" \
                "Please either create the directory or change the name." \
                % (outputDirectory),
                "Output Directory Does Not Exist",
                    wx.OK | wx.ICON_ERROR)
            msgOutputDirectoryDoesNotExistDlg.ShowModal()
            msgOutputDirectoryDoesNotExistDlg.Destroy()
            return False   

        #Check whether contains problematic characters
        if outputDirectory.find("'") > -1 or outputDirectory.find("\"") > -1 or outputDirectory.find("$") > -1 :

            #If so, give option of going back or trying anyway
            #NB, it is not just Python files that need fixing. 
            #Many lines throughout makemenu do aswell.
            msgOutputDirectoryErrorDlg = wx.MessageDialog(panel, \
                "For technical reasons, currently apostrophes, double quotes\n"\
                "and $ signs may cause problems when in the output directory name.\n" \
                "The output directory name is: %s\n\n" \
                "Do you want to return to rename the output directory?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (outputDirectory),
                "Problematic Character in Output Directory",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgOutputDirectoryErrorDlg.ShowModal()
            msgOutputDirectoryErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        #If get this far, file is OK
        return True

    def RelevantFilesAreOK(self, panel):
        """Check the disc options for errors detectable before encoding"""
        if self.OutputDirectoryExists(panel) == False:
            return False
        if self.DiscNameOK(panel) == False:
            return False
        return True

class GroupOptions:
    """Options related to encoding a group"""
    # Type of item being encoded (menu, video, group or slides)
    type = ID_GROUP
    # Title of the group itself
    title = "Untitled group"
    # Additional command-line options
#    addoptions = ""
    # Duration is unknown
    #duration = -1 
    inFile = ""
    outPrefix = ""
    groupMemberCount = 0

    def __init__(self, filename = ""):
        self.inFile = filename
        self.title = os.path.basename(filename).replace('_', ' ')
        self.outPrefix = "%s.tovid_encoded" % self.title

    def toElement(self):
        """Convert the current options into a Group and return it."""
        # Get global configuration (for output directory)
        curConfig = TovidConfig()
        # Create Group and set relevant options
        group = Group(self.title)
        group['tvsys'] = self.tvsys
        group['format'] = self.format
        group['aspect'] = self.aspect
        group['in'] = self.inFile
        group['group'] = "%s/%s" % (curConfig.strOutputDirectory, self.outPrefix)
        return group

    def fromElement(self, group):
        """Load options from a Group."""
        print "Loading Group:"
        print group.to_string()
        self.type = ID_GROUP
        self.inFile = group['in']
        self.outPrefix = group['out']


    def GetCommand(self):
        """Return the complete tovid command for converting this group."""
        #No command is required for groups
        return ""

    def SetDiscFormat(self, format):
        """Make group compliant with given disc format."""
        #Placeholder        

    def SetDiscTVSystem(self, tvsys):
        """Make group compliant with given disc TV system."""
        #Placeholder

    def CopyFrom(self, opts):
        """Copy the given options into this object."""
        # If types are different, return
        if self.type != opts.type:
            return

    def GroupOutputOK(self, panel):
        """Check the group output for any errors detectable before encoding"""
        # Get global configuration (for output directory)
        curConfig = TovidConfig()

        groupOutput = "%s" % (self.outPrefix)

        #Check whether it is none null (as this gives an error)
        if self.outPrefix == "":
            msgDiscFileExistsDlg = wx.MessageDialog(panel, \
                    "Please enter a valid name (i.e. not empty) for each Group.", 
                    "Missing Group Name",
                    wx.OK | wx.ICON_ERROR)
            msgDiscFileExistsDlg.ShowModal()
            msgDiscFileExistsDlg.Destroy()
            return False

        #Check whether the output file contains problematic characters
        if groupOutput.find("'") > -1 or groupOutput.find("\"") > -1 or groupOutput.find("$") > -1 :

            #If so, give option of going back or trying anyway
            #NB, it is not just Python files that need fixing. 
            #Many lines throughout e.g. makemenu do aswell.
            msgGroupOutputErrorDlg = wx.MessageDialog(panel, \
                "This program converts the group files into the correct format,\n" \
                "and saves these resultant files with names based on the entered values.\n\n" \
                "For technical reasons, currently apostrophes, double quotes\n" \
                "and $ signs may cause problems when in these filenames.\n" \
                "One output file contains at least one of these characters.\n"
                "This output group file is: %s\n\n" \
                "Do you want to return to modify these values?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (groupOutput),
                "Problematic character in Group Output Filename",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgGroupOutputErrorDlg.ShowModal()
            msgGroupOutputErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        return True

    def RelevantFilesAreOK(self, panel):
        """Check the group options for any errors detectable before encoding"""
        if self.GroupOutputOK(panel) == False:
            return False
        if self.groupMemberCount == 0:
            msgEmptyGroupDlg = wx.MessageDialog(panel, \
                    "Each Group must have at least one video.", 
                    "Empty Group",
                    wx.OK | wx.ICON_ERROR)
            msgEmptyGroupDlg.ShowModal()
            msgEmptyGroupDlg.Destroy()
            return False
        return True

class MenuOptions:
    """Options related to generating a menu"""
    # Type of item being encoded (menu, video, or slides)
    type = ID_MENU
    isTopMenu = False
    # Title of the menu itself
    title = _("Untitled menu")
    # -ntsc or -pal
    tvsys = 'ntsc'
    # -dvd, -vcd, or -svcd
    format = 'dvd'
    # -menu-title
    menutitle = ""
    titlefontsize = "32"
    # -background FILE
    background = ""
    # -audio FILE
    audio = ""
    # -length NUM
    menulength = ""
    # -align [northwest|north|northeast]
    alignment = 'northwest'
    # Menu text settings
    fontsize = "24"
    fillType = "Color"
    fillColor1 = "rgb(255,255,255)"      # For fill, gradient, fractals
    fillColor2 = "rgb(255,255,255)"      # For gradient and fractals
    color1 = wx.Colour(255, 255, 255)    # Stores dialog box color
    color2 = wx.Colour(255, 255, 255)    # Stores dialog box color
    textStrokeColor = "rgb(0,0,0)"       # For text outline
    colorStroke = wx.Colour(0, 0, 0)     # Stores dialog box color
    textStrokeWidth = "1"
    pattern = "bricks"
    # DVD button settings
    button = ">"
    highlightColor = "rgb(255,255,0)"              # For highlight color
    selectColor = "rgb(255,0,0)"                   # for selection color
    colorHi = wx.Colour(255, 255, 0)               # Stores dialog box color
    colorSel = wx.Colour(255, 0, 0)                # Stores dialog box color
    buttonOutlineColor = "rgb(140,140,140)"        # For button outline
    colorButtonOutline = wx.Colour(140, 140, 140)  # Stores dialog box color
    # Other settings
    titles = []
    outPrefix = ""

    def __init__(self, format = 'dvd', tvsys = 'ntsc',
        title = _("Untitled menu"), isTop = False):
        self.SetDiscFormat(format)
        self.SetDiscTVSystem(tvsys)
        self.title = title
        self.outPrefix = title.replace(' ', '_')
        self.isTopMenu = isTop
        # Default font
        self.font = wx.Font(12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
            wx.FONTWEIGHT_NORMAL, False, "Default")
        self.buttonFont = wx.Font(12, wx.FONTFAMILY_DEFAULT, 
            wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False, "Default")

    def GetCommand(self):
        """Return the complete makemenu command to generate this menu."""
        # Get global configuration (for output directory)
        curConfig = TovidConfig()

        strCommand = "makemenu -noask -quiet -overwrite "
        # Append format and tvsys
        strCommand += "-%s -%s " % (self.tvsys, self.format)
        # Append alignment
        strCommand += "-align %s " % self.alignment

        # Image and audio backgrounds, if any
        if self.menutitle != "":
            strCommand += "-menu-title \"%s\" " % self.menutitle
            strCommand += "-menu-title-fontsize %s " % self.titlefontsize
        if self.background != "":
            strCommand += "-background \"%s\" " % self.background
        if self.audio != "":
            strCommand += "-audio \"%s\" " % self.audio
        if self.menulength != "":
            strCommand += "-length %s " % self.menulength

        # Append font and size
        if self.font.GetFaceName() != "":
            wx_FontName = self.font.GetFaceName()
            IM_FontName = curConfig.wx_IM_FontMap[wx_FontName] 
            strCommand += "-font \"%s\" " % IM_FontName
        strCommand += "-fontsize %s " % self.fontsize

        # Express text fill and outline in ways makemenu will understand
        # Text outline:
        fontDecoration = "-stroke %s" % (self.textStrokeColor)
        fontDecoration += " -strokewidth %s" % (self.textStrokeWidth)
        # Text fill:
        # solid is simple:
        if self.fillType == "Color":
            strCommand += "-textcolor \"%s\" " % self.fillColor1
        # fractals, gradients, and patterns get put in -fontdeco
        elif self.fillType == "Fractal":
            fontDecoration += " -tile fractal:%s-%s" % \
                (self.fillColor1, self.fillColor2)
        elif self.fillType == "Gradient":
            fontDecoration += " -tile gradient:%s-%s" % \
                (self.fillColor1, self.fillColor2)
        elif self.fillType == "Pattern":
            fontDecoration += " -tile pattern:%s" % (self.pattern)
        else:
            pass
        
        strCommand += "-fontdeco '%s' " % fontDecoration

        # For DVD, add button options
        if self.format == 'dvd':
            # Font
            if self.buttonFont.GetFaceName() != "":
                wx_FontName = self.buttonFont.GetFaceName()
                IM_FontName = curConfig.wx_IM_FontMap[wx_FontName] 
                # Don't need button font for 'play' or 'movie', take the rest
                if self.button != "play" and self.button != "movie":
                    strCommand += "-button-font \"%s\" " % IM_FontName
            # Shape
            strCommand += "-button \"%s\" " % self.button
            # colors
            strCommand += "-highlightcolor \"%s\" " % self.highlightColor
            strCommand += "-selectcolor \"%s\" " % self.selectColor
            strCommand += "-button-outline \"%s\" " % self.buttonOutlineColor

        # Append video/still titles
        for title in range(len(self.titles)):
            strCommand += "\"%s\" " % self.titles[ title ]

        strCommand += "-out \"%s/%s\"" % \
                (curConfig.strOutputDirectory, self.outPrefix)
        return strCommand

    def SetDiscFormat(self, format):
        """Make menu compliant with given disc format."""
        self.format = format 

    def SetDiscTVSystem(self, tvsys):
        """Make menu compliant with given disc TV system."""
        self.tvsys = tvsys

    def CopyFrom(self, opts):
        """Copy the given MenuOptions' data into this menu."""
        # Return if types are different
        if self.type != opts.type:
            return
        # Copy opts into self
        self.format = opts.format
        self.tvsys = opts.tvsys
        self.background = opts.background
        self.audio = opts.audio
        self.font = opts.font
        self.alignment = opts.alignment
        self.menutitle = opts.menutitle
        self.titlefontsize = opts.titlefontsize
        self.menulength = opts.menulength
        self.fontsize = opts.fontsize
        self.fillType = opts.fillType
        self.fillColor1 = opts.fillColor1
        self.color1 = opts.color1
        self.fillColor2 = opts.fillColor2
        self.color2 = opts.color2
        self.textStrokeColor = opts.textStrokeColor
        self.colorStroke = opts.colorStroke
        self.textStrokeWidth = opts.textStrokeWidth
        self.buttonFont = opts.buttonFont
        self.highlightColor = opts.highlightColor
        self.colorHi = opts.colorHi
        self.selectColor = opts.selectColor
        self.colorSel = opts.colorSel
        self.button = opts.button
        self.buttonOutlineColor = opts.buttonOutlineColor
        self.colorButtonOutline = opts.colorButtonOutline

    def AudioFileOK(self, panel):
        """Check for any errors associated with the audio file"""
        #First check if Audio File is specified
        audioFile = self.audio
        if audioFile == "":
            return True
        
        #If specified, check whether it exists
        if os.path.isfile(audioFile) == False:
            msgAudioFileMissingDlg = wx.MessageDialog(panel, \
                    "A menu has an audio file that does not appear to exist.\n" \
                    "The file is: %s\n\n" \
                    "Please choose another audio file." % (audioFile),
                    "Missing Audio File",
                    wx.OK | wx.ICON_ERROR)
            msgAudioFileMissingDlg.ShowModal()
            msgAudioFileMissingDlg.Destroy()
            return False

        #If exists, check whether contains problematic characters
        if audioFile.find("'") > -1 or audioFile.find("\"") > -1 or audioFile.find("$") > -1 :

            #If so, give option of going back or trying anyway
            #NB, it is not just Python files that need fixing. 
            #Many lines throughout e.g. makemenu do aswell.
            msgAudioErrorDlg = wx.MessageDialog(panel, \
                "For technical reasons, currently apostrophes, double quotes\n" \
                "and $ signs may cause problems in files associated with menus.\n" \
                "One audio file contains at least one of these characters.\n"
                "This audio file is: %s\n\n" \
                "Do you want to return to rename / choose another audio file?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (audioFile),
                "Problematic character in Audio File",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgAudioErrorDlg.ShowModal()
            msgAudioErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        #If get this far, file is OK
        return True

    def ImageFileOK(self, panel):
        """Check for any errors associated with the image file"""
        #First check if Image File is specified
        imageFile = self.background
        if imageFile == "":
            return True
        
        #If specified, check whether it exists
        if os.path.isfile(imageFile) == False:
            msgImageFileMissingDlg = wx.MessageDialog(panel, \
                    "A menu has an image file that does not appear to exist.\n" \
                    "The file is: %s\n\n" \
                    "Please choose another image file." % (imageFile),
                    "Missing Image File",
                    wx.OK | wx.ICON_ERROR)
            msgImageFileMissingDlg.ShowModal()
            msgImageFileMissingDlg.Destroy()
            return False

        #If exists, check whether contains problematic characters
        if imageFile.find("'") > -1 or imageFile.find("\"") > -1 or imageFile.find("$") > -1 :

            #If so, give option of going back or trying anyway
            #NB, it is not just Python files that need fixing. 
            #Many lines throughout e.g. makemenu do aswell.
            msgImageErrorDlg = wx.MessageDialog(panel, \
                "For technical reasons, currently apostrophes, double quotes\n" \
                "and $ signs may cause problems in files associated with menus.\n" \
                "One image file contains at least one of these characters.\n"
                "This image file is: %s\n\n" \
                "Do you want to return to rename / choose another image file?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (imageFile),
                "Problematic character in Image File",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgImageErrorDlg.ShowModal()
            msgImageErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        #If get this far, file is OK
        return True

    def HeaderOK(self, panel):
        """Check for any errors associated with the menu header"""
        menuTitle = "%s" % (self.menutitle)

        #Check whether menu title contains problematic characters
        if menuTitle.find("'") > -1 or menuTitle.find("\"") > -1 \
           or menuTitle.find("$") > -1:

            #If so, give option of going back or trying anyway
            msgMenuTitleErrorDlg = wx.MessageDialog(panel, \
                "For technical reasons, currently apostrophes, double quotes\n" \
                "and $ signs may cause problems in the menu headers.\n" \
                "One menu header contains at least one of these characters.\n"
                "This menu header is: %s\n\n" \
                "Do you want to return to rename this menu header?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (menuTitle),
                "Problematic character in Menu Title",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgMenuTitleErrorDlg.ShowModal()
            msgMenuTitleErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        #If get this far, file is OK
        return True

    def OutputFileOK(self, panel):
        """Check for any errors associated with the output file"""
        # Get global configuration (for output directory)
        curConfig = TovidConfig()

        #First check if Image File is specified
        outputFile = "%s/%s.mpg" % (curConfig.strOutputDirectory, self.outPrefix)
        
        #If specified, check whether it exists
        if os.path.exists(outputFile) == True:
            #If exists, check whether it is a file or directory
            if os.path.isfile(outputFile) == True:
                msgOutputFileMissingDlg = wx.MessageDialog(panel, \
                    "A file with the same name as a menu output file already exists.\n" \
                    "This file is: %s\n\n" \
                    "Do you want to continue?\n"
                    "This will overwrite this existing file." % (outputFile),
                    "Existing Menu Output File",
                    wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
                response = msgOutputFileMissingDlg.ShowModal()
                msgOutputFileMissingDlg.Destroy()
                if response != wx.ID_YES:
                    return False   
            else:   
                msgOutputFolderExistsDlg = wx.MessageDialog(panel, \
                    "A file needs to be created based on a menu name.\n" \
                    "But, a directory with the same name already exists.\n" \
                    "This directory is: %s\n\n" \
                    "Please either rename/remove the directory or " \
                    "change the menu name!" % (outputFile),
                    "Existing Directory",
                    wx.OK | wx.ICON_ERROR)
                msgOutputFolderExistsDlg.ShowModal()
                msgOutputFolderExistsDlg.Destroy()
                return False   


        #Check whether output file contains problematic characters
        if outputFile.find("'") > -1 or outputFile.find("\"") > -1 \
           or outputFile.find("$") > -1 or outputFile.find("&") > -1:

            #If so, give option of going back or trying anyway
            #NB, it is not just Python files that need fixing. 
            #Many lines throughout e.g. makemenu do aswell.
            msgMenuNameErrorDlg = wx.MessageDialog(panel, \
                "For technical reasons, currently apostrophes, double quotes\n" \
                "& signs and $ signs may cause problems in the name of menus.\n" \
                "One menu name contains at least one of these characters.\n"
                "This menu name is: %s\n\n" \
                "Do you want to return to rename this menu?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (outputFile),
                "Problematic character in Menu Name",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgMenuNameErrorDlg.ShowModal()
            msgMenuNameErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        #If get this far, file is OK
        return True

    def RelevantFilesAreOK(self, panel):
        """Check the menu options for any errors detectable before encoding"""
        if self.AudioFileOK(panel) == False:
            return False
        if self.ImageFileOK(panel) == False:
            return False
        if self.HeaderOK(panel) == False:
            return False
        if self.OutputFileOK(panel) == False:
            return False
        return True


class SlideOptions:
    """Options related to generating a slideshow"""
    # Type of item being encoded (menu, video, or slides)
    type = ID_SLIDE
    # Title of the group of slides
    title = _("Untitled slides")
    # List of image files to convert to slides
    files = []
    # -dvd, -vcd, or -svcd
    format = 'dvd' 
    # -ntsc or -pal
    tvsys = 'ntsc'

    def __init__(self, format = 'dvd', tvsys = 'ntsc',
        files = []):
        self.tvsys = tvsys
        self.format = format 
        self.files = files

    def GetCommand(self):
        """Return the makeslides command to generate this slideshow."""
        strCommand = "makeslides -%s -%s " % \
            (self.tvsys, self.format)

    def SetDiscFormat(self, format):
        """Make slides compliant with given disc format."""
        self.format = format 

    def SetDiscTVSystem(self, tvsys):
        """Make slides compliant with given disc TV system."""
        self.tvsys = tvsys

    def CopyFrom(self, opts):
        """Copy the given options into this object."""
        # If types are different, return
        if self.type != opts.type:
            return
        # Copy opts into self
        self.format = opts.format
        self.tvsys = opts.tvsys

    def RelevantFilesAreOK(self, panel):
        """Check the slide options for any errors detectable before encoding"""
        return True

class VideoOptions:
    """Options related to encoding a video"""
    # Type of item being encoded (menu, video or slides)
    type = ID_VIDEO
    # Title of the video itself
    title = "Untitled video"
    # -dvd, -half-dvd, -svcd, -dvd-vcd, -vcd
    format = 'dvd'
    # -ntsc or -pal
    tvsys = 'ntsc'
    # -full, -wide, or -panavision
    aspect = 'full'
    # string 1 to 10
    quality = "8"
    # mpeg2enc or ffmpeg
    encodingMethod = "mpeg2enc"
    # Additional command-line options
    addoptions = ""
    # Duration is unknown
    #duration = -1 
    inFile = ""
    outPrefix = ""

    def __init__(self, format = 'dvd', tvsys = 'ntsc',
        filename = ""):
        self.SetDiscTVSystem(tvsys)
        self.SetDiscFormat(format)
        self.inFile = filename
        self.title = os.path.basename(filename).replace('_', ' ')
        self.outPrefix = "%s.tovid_encoded" % self.title

    def GetCommand(self):
        """Return the complete tovid command for converting this video."""
        # Get global configuration (for output directory)
        curConfig = TovidConfig()

        # Always overwrite. Use better solution in future?
        strCommand = "tovid -quiet -overwrite "
        # Append tvsys, format, and aspect ratio
        strCommand += "-%s -%s -%s " % \
            (self.tvsys, self.format, self.aspect)

        # Append quality
        strCommand += "-quality %s " % self.quality

        # Append encoding method if not default
        if self.encodingMethod == "ffmpeg":
            strCommand += "-ffmpeg "
 
        # Append other options
        if self.addoptions:
            strCommand += "%s " % self.addoptions

        strCommand += "-in \"%s\" " % self.inFile
        strCommand += "-out \"%s/%s\" " % (curConfig.strOutputDirectory, self.outPrefix)
        strCommand += "-from-gui -noask"
        return strCommand

    def SetDiscFormat(self, format):
        """Make video compliant with given disc format."""
        if format == 'vcd':
            self.format = format
        elif format == 'svcd':
            self.format = format
        elif format == 'dvd':
            # Don't change existing format unless it's VCD or SVCD
            if self.format in [ 'vcd', 'svcd' ]:
                self.format = 'dvd'

    def SetDiscTVSystem(self, tvsys):
        """Make menu compliant with given disc TV system."""
        self.tvsys = tvsys

    def CopyFrom(self, opts):
        """Copy the given options into this object."""
        # If types are different, return
        if self.type != opts.type:
            return
        # Copy opts into self
        self.format = opts.format
        self.tvsys = opts.tvsys
        self.aspect = opts.aspect
        self.quality = opts.quality
        self.encodingMethod = opts.encodingMethod
        self.addoptions = opts.addoptions

    def VideoFileOK(self, panel):
        """Check the video options for any errors detectable before encoding"""
        #First check if Video File is specified
        videoFile = self.inFile
        
        #Check whether it exists
        if os.path.isfile(videoFile) == False:
            msgVideoFileMissingDlg = wx.MessageDialog(panel, \
                    "A Video does not appear to exist.\n" \
                    "The file is: %s\n\n" \
                    "Please choose another video file." % (videoFile),
                    "Missing Video File",
                    wx.OK | wx.ICON_ERROR)
            msgVideoFileMissingDlg.ShowModal()
            msgVideoFileMissingDlg.Destroy()
            return False

        #If exists, check whether contains problematic characters
        if videoFile.find("'") > -1 or \
           videoFile.find("\"") > -1 or \
           videoFile.find("$") > -1 :

            #If so, give option of going back or trying anyway
            msgVideoErrorDlg = wx.MessageDialog(panel, \
                "For technical reasons, currently apostrophes, double quotes\n" \
                "and $ signs may cause problems in the file name of videos.\n" \
                "One video file contains at least one of these characters.\n"
                "This video file is: %s\n\n" \
                "Do you want to return and rename / choose another video file?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (videoFile),
                "Problematic character in Video File",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgVideoErrorDlg.ShowModal()
            msgVideoErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        return True

    def VideoOutputOK(self, panel):
        """Check the video output for any errors detectable before encoding"""
        # Get global configuration (for output directory)
        curConfig = TovidConfig()

        videoOutput = "%s/%s.mpg" % (curConfig.strOutputDirectory, self.outPrefix)

        #Check whether it is none null (as this gives an error)
        if self.outPrefix == "":
            msgDiscFileExistsDlg = wx.MessageDialog(panel, \
                    "Please enter a valid name (i.e. not empty) for each Video.", 
                    "Missing Video Name",
                    wx.OK | wx.ICON_ERROR)
            msgDiscFileExistsDlg.ShowModal()
            msgDiscFileExistsDlg.Destroy()
            return False

        #Check whether output already exists
        #If specified, check whether it exists
        if os.path.exists(videoOutput) == True:
            #If exists, check whether it is a file or directory
            if os.path.isfile(videoOutput) == True:
                msgVideoOutputExistsErrorDlg = wx.MessageDialog(panel, \
                    "This program converts the video files into the correct format,\n" \
                    "and saves these resultant files with names based on the entered values.\n\n" \
                    "One of these output files already exists.\n" \
                    "This output file is: %s\n\n" \
                    "Do you want to overwrite this file?" % (videoOutput),
                    "Problematic character in Video Output Filename",
                    wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
                response = msgVideoOutputExistsErrorDlg.ShowModal()
                msgVideoOutputExistsErrorDlg.Destroy()
       
                if response != wx.ID_YES:
                    return False   
            else:   
                msgVideoOutputFolderExistsDlg = wx.MessageDialog(panel, \
                    "A file needs to be created based on the video name.\n" \
                    "But, a directory with the same name already exists.\n" \
                    "This directory is: %s\n\n" \
                    "Please either rename/remove the directory or " \
                    "change the menu name!" % (videoOutput),
                    "Existing Directory",
                    wx.OK | wx.ICON_ERROR)
                msgVideoOutputFolderExistsDlg.ShowModal()
                msgVideoOutputFolderExistsDlg.Destroy()
                return False   

        #Check whether the output file contains problematic characters
        if videoOutput.find("'") > -1 or videoOutput.find("\"") > -1 \
           or videoOutput.find("$") > -1 or videoOutput.find("&") > -1:

            #If so, give option of going back or trying anyway
            #NB, it is not just Python files that need fixing. 
            #Many lines throughout e.g. makemenu do aswell.
            msgVideoOutputErrorDlg = wx.MessageDialog(panel, \
                "This program converts the video files into the correct format,\n" \
                "and saves these resultant files with names based on the entered values.\n\n" \
                "For technical reasons, currently apostrophes, double quotes\n" \
                "& and $ signs may cause problems when in these filenames.\n" \
                "One output file contains at least one of these characters.\n"
                "This output video file is: %s\n\n" \
                "Do you want to return to modify these values?\n" \
                "NB, continuing (i.e. No) is very unlikely to work!" % (videoOutput),
                "Problematic character in Video Output Filename",
                wx.YES_NO | wx.YES_DEFAULT | wx.ICON_ERROR)
            response = msgVideoOutputErrorDlg.ShowModal()
            msgVideoOutputErrorDlg.Destroy()
       
            if response != wx.ID_NO:
                return False   
        return True

    def RelevantFilesAreOK(self, panel):
        """Check the video options for any errors detectable before encoding"""
        if self.VideoFileOK(panel) == False:
            return False
        if self.VideoOutputOK(panel) == False:
            return False
        return True



#! /usr/bin/env python
# frames.py

import os
import wx

import libtovid
from libtovid.gui.configs import TovidConfig
from libtovid.gui.constants import *
from libtovid.gui.icons import AppIcon
from libtovid.gui.panels import *
from libtovid.gui.util import _

__all__ = [\
    "TovidFrame",
    "MiniEditorFrame",
    "TodiscFrame"]

class TovidFrame(wx.Frame):
    """Main tovid GUI frame. Contains and manages all sub-panels.
    """

    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id , title, wx.DefaultPosition,
            (800, 820), wx.DEFAULT_FRAME_STYLE | wx.NO_FULL_REPAINT_ON_RESIZE)

        icon = wx.EmptyIcon()
        icon.CopyFromBitmap(AppIcon())
        self.SetIcon(icon)

        # Global configuration
        self.curConfig = TovidConfig()

        self.dirname = os.getcwd()

        # Menu bar
        self.menubar = wx.MenuBar()

        # File menu
        self.menuFile = wx.Menu()
        #self.menuFile.Append(ID_MENU_FILE_PREFS, "&Preferences",
        #    "Configuration settings for tovid GUI")
        #self.menuFile.AppendSeparator()
        self.menuFile.Append(ID_MENU_FILE_EXIT, "E&xit",
                "Exit tovid GUI")

        # Help menu
        self.menuHelp = wx.Menu()
        self.menuHelp.Append(ID_MENU_HELP_ABOUT, "&About",
            "Information about this program")

        # View menu
        self.menuView = wx.Menu()
        # Toggle options
        self.menuViewShowGuide = wx.MenuItem(self.menuView, ID_MENU_VIEW_SHOWGUIDE,
            "Show &guide", "Show/hide the tovid guide panel", wx.ITEM_CHECK)
        self.menuViewShowTooltips = wx.MenuItem(self.menuView, ID_MENU_VIEW_SHOWTOOLTIPS,
            "Show &tooltips", "Show/hide tooltips in the GUI", wx.ITEM_CHECK)
        # Add items to View menu
        self.menuView.AppendItem(self.menuViewShowGuide)
        self.menuView.AppendItem(self.menuViewShowTooltips)
        self.menuViewShowGuide.Check(False)
        self.menuViewShowTooltips.Check(True)

        #self.menuLang = wx.Menu()
        #self.menuLang.Append(ID_MENU_LANG_EN, "&English")
        #self.menuLang.Append(ID_MENU_LANG_DE, "&Deutsch")
        # Append language menu as a submenu of View
        #self.menuView.AppendMenu(ID_MENU_LANG, "Language", self.menuLang)

        
        # Menu events
        # TODO: Re-enable save/open
        #wx.EVT_MENU(self, ID_MENU_FILE_PREFS, self.OnFilePrefs)
        wx.EVT_MENU(self, ID_MENU_FILE_EXIT, self.OnExit)
        wx.EVT_MENU(self, ID_MENU_VIEW_SHOWGUIDE, self.OnShowGuide)
        wx.EVT_MENU(self, ID_MENU_VIEW_SHOWTOOLTIPS, self.OnShowTooltips)
        wx.EVT_MENU(self, ID_MENU_HELP_ABOUT, self.OnAbout)
        wx.EVT_MENU(self, ID_MENU_LANG_EN, self.OnLang)
        wx.EVT_MENU(self, ID_MENU_LANG_DE, self.OnLang)
        self.Bind(wx.EVT_KEY_UP, self.OnKeyUp)

        # Build menubar
        self.menubar.Append(self.menuFile, "&File")
        self.menubar.Append(self.menuView, "&View")
        self.menubar.Append(self.menuHelp, "&Help")
        self.SetMenuBar(self.menubar)

        # Toolbar
        #self.toolbar = self.CreateToolBar(wx.TB_HORIZONTAL)
        #self.toolbar.AddControl(wx.ContextHelpButton(self.toolbar))
        #self.toolbar.Realize()

        # Statusbar
        self.curConfig.statusBar = self.CreateStatusBar();
        self.Show(True)

        # Guide panel
        self.panGuide = GuidePanel(self, wx.ID_ANY)
        self.panGuide.Hide()
        # Store guide panel reference in global config
        self.curConfig.panGuide = self.panGuide
        
        # Task panels. AuthorFiles is currently the only working task.
        self.panAuthorFiles = AuthorFilesTaskPanel(self, wx.ID_ANY)

        # Task sizer. Holds current 3-step task
        self.sizTask = wx.BoxSizer(wx.VERTICAL)
        self.sizTask.Add(self.panAuthorFiles, 1, wx.EXPAND)

        # Main sizer. Holds task (layout, encoding) panel and Guide panel
        self.sizMain = wx.BoxSizer(wx.HORIZONTAL)
        #self.sizMain.Add(self.panGuide, 2, wx.EXPAND | wx.ALL, 8)
        #self.sizMain.Add(wx.StaticLine(self, wx.ID_ANY, style = wx.LI_VERTICAL),
        #    0, wx.EXPAND)
        self.sizMain.Add(self.sizTask, 5, wx.EXPAND)
        self.SetSizer(self.sizMain)

    def OnExit(self, evt):
        """Exit the GUI and close all windows."""
        self.Close(True)

    def OnShowGuide(self, evt):
        """Show or hide the guide panel."""
        if evt.IsChecked():
            self.sizMain.Prepend(self.panGuide, 2, wx.EXPAND | wx.ALL, 8)
            self.panGuide.Show()
            self.sizMain.Layout()
        else:
            self.sizMain.Remove(0)
            self.panGuide.Hide()
            self.sizMain.Layout()

    def OnShowTooltips(self, evt):
        """Show or hide GUI tooltips."""
        if evt.IsChecked():
            # Enable tooltips globally
            wx.ToolTip_Enable(True)
        else:
            # Disable tooltips globally
            wx.ToolTip_Enable(False)
        
    def OnAbout(self, evt):
        """Display a dialog showing information about tovidgui."""
        strAbout = "You are using the tovid GUI, version 0.31,\n" \
          "part of the tovid video disc authoring suite.\n\n" \
          "For more information and documentation, please\n" \
          "visit the tovid web site:\n\n" \
          "http://tovid.org/"
        dlgAbout = wx.MessageDialog(self, strAbout, "About tovid GUI", wx.OK)
        dlgAbout.ShowModal()

    def OnLang(self, evt):
        """Change GUI to selected language."""
        if evt.GetId() == ID_MENU_LANG_EN:
            self.curConfig.UseLanguage('en')
            
        elif evt.GetId() == ID_MENU_LANG_DE:
            self.curConfig.UseLanguage('de')

    def OnKeyUp(self, evt):
        """Key up event handler.  Primarily used to close the app if certain keys are pressed.
        """
        key = evt.KeyCode()
        if (key >= 0 and key < 256):
            controlDown = evt.ControlDown()
            if controlDown and "Q" == chr(key):
                self.Close()

    #def OnFilePrefs(self, evt):
    #    """Open preferences window and set configuration"""
    #    dlgPrefs = PreferencesDialog(self, wx.ID_ANY)
    #    dlgPrefs.ShowModal()
    #    # Set the output directory in the PreEncoding panel
    #    self.panEncoding.SetOutputDirectory(self.curConfig.strOutputDirectory)
# end of class TovidFrame          

class MiniEditorFrame(wx.Frame):
    """Simple text editor (for editing configuration files) in a frame"""
    def __init__(self, parent, id, filename):
        wx.Frame.__init__(self, parent, id, "%s (MiniEditor)" % filename, \
            wx.DefaultPosition, (500, 400), \
            wx.DEFAULT_FRAME_STYLE | wx.NO_FULL_REPAINT_ON_RESIZE)

        # Current directory is current dirname
        self.dirname = os.getcwd()
        # Center dialog
        self.Centre()
        # Menu bar
        self.menubar = wx.MenuBar()
        # File menu
        self.menuFile = wx.Menu()
        self.menuFile.Append(ID_MENU_FILE_NEW, "&New", "Start a new file")
        self.menuFile.AppendSeparator()
        self.menuFile.Append(ID_MENU_FILE_OPEN, "&Open", "Open a new file")
        self.menuFile.Append(ID_MENU_FILE_SAVE, "&Save", "Save current file")
        self.menuFile.AppendSeparator()
        self.menuFile.Append(ID_MENU_FILE_EXIT, "E&xit", "Exit mini editor")
        # Menu events
        wx.EVT_MENU(self, ID_MENU_FILE_NEW, self.OnNew)
        wx.EVT_MENU(self, ID_MENU_FILE_OPEN, self.OnOpen)
        wx.EVT_MENU(self, ID_MENU_FILE_SAVE, self.OnSave)
        wx.EVT_MENU(self, ID_MENU_FILE_EXIT, self.OnExit)
        # Build menubar
        self.menubar.Append(self.menuFile, "&File")
        self.SetMenuBar(self.menubar)
        # Editor window
        self.editWindow = wx.Editor(self, wx.ID_ANY, style = wx.SUNKEN_BORDER)
        self.sizMain = wx.BoxSizer(wx.VERTICAL)
        self.sizMain.Add(self.editWindow, 1, wx.EXPAND | wx.ALL, 6)
        # Open the given filename and load its text into the editor window
        self.OpenFile(filename)
        self.SetSizer(self.sizMain)

    def OnNew(self, evt):
        """Edit a new file"""
        # Just empty edit window and set filename to null
        self.editWindow.SetText([""])
        self.currentFilename = ""
        # Set title to indicate that a new file is being edited
        self.SetTitle("Untitled file (MiniEditor)")

    def OnOpen(self, evt):
        inFileDialog = wx.FileDialog(self, _("Choose a file to open"), self.dirname, "",
            "*.*", wx.OPEN)
        if inFileDialog.ShowModal() == wx.ID_OK:
            self.dirname = inFileDialog.GetDirectory()
            # Open the file
            self.OpenFile(inFileDialog.GetPath())

        inFileDialog.Destroy()

    def OnSave(self, evt):
        """Save the current file."""
        if self.currentFilename:
            # Save the current filename (prompt for overwrite)
            confirmDialog = wx.MessageDialog(self, \
                    _("Overwrite the existing file?"), \
                    _("Confirm overwrite"), wx.YES_NO)
            # If not overwriting, return now
            if confirmDialog.ShowModal() == wx.ID_NO:
                return

        # currentFilename is empty; prompt for a filename
        else:
            outFileDialog = wx.FileDialog(self, \
                    _("Choose a filename to save"), \
                    self.dirname, "", "*.*", wx.SAVE)
            if outFileDialog.ShowModal() == wx.ID_OK:
                self.dirname = outFileDialog.GetDirectory()
                self.currentFilename = outFileDialog.GetPath()

        outFile = open(self.currentFilename, "w")
        for line in self.editWindow.GetText():
            outFile.write("%s\n" % line)
        outFile.close()
        # Update title bar
        self.SetTitle("%s (MiniEditor)" % self.currentFilename)

    def OnExit(self, evt):
        self.Close(True)

    def OpenFile(self, filename):
        """Open the given filename in the editor."""
        # Save the filename locally
        self.currentFilename = filename
        # Create the file input stream
        inFile = open(filename, 'r')
        buffer = inFile.readline()
        strEditText = []
        # Read from the file until EOF
        while buffer:
            buffer = inFile.readline()
            # Append file's text to a string
            strEditText.append(buffer)
        # Fill the editor window with the text buffer
        self.editWindow.SetText(strEditText)
        inFile.close()
        # Update titlebar text
        self.SetTitle("%s (MiniEditor)" % self.currentFilename)
# end of class MiniEditorFrame

class TodiscFrame(wx.Frame):
    """Main todisc GUI frame. Contains and manages all sub-panels.
    """
    def __init__(self, *args, **kwds):
        kwds["style"] = wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, *args, **kwds)
        self.SetTitle(_("todisc GUI"))
        
        _icon = wx.EmptyIcon()
        _icon.CopyFromBitmap(AppIcon())
        self.SetIcon(_icon)
        self.SetFocus()
        #
        # Create the menu bar
        #
        self.mbMain = wx.MenuBar()
        self.SetMenuBar(self.mbMain)
        mnFile = wx.Menu()
        mnFile.Append(ID_MENU_FILE_EXIT, _("E&xit"), "mnExit", wx.ITEM_NORMAL)
        self.mbMain.Append(mnFile, _("&File"))
        mnView = wx.Menu()
        mnView.Append(ID_MENU_VIEW_ADVANCED, _("&Advanced"), "mnAdvanced", wx.ITEM_CHECK)
        self.mbMain.Append(mnView, _("&View"))
        mnHelp = wx.Menu()
        mnHelp.Append(ID_MENU_HELP_ABOUT, _("&About"), "mnAbout", wx.ITEM_NORMAL)
        self.mbMain.Append(mnHelp, _("&Help"))
        #
        # Create the notebook for tabs
        #
        self.notebook_1 = wx.Notebook(self, wx.ID_ANY, style=0)
        self.nbPlaylist = PlaylistTabPanel(self.notebook_1, wx.ID_ANY)
        self.nbMenus = MenuTabPanel(self.notebook_1, wx.ID_ANY)
        self.nbThumbnails = ThumbnailTabPanel(self.notebook_1, wx.ID_ANY)
        self.nbDebug = DebugTabPanel(self.notebook_1, wx.ID_ANY)
        self.nbDebug.Hide()
        #
        # Add pages to notebook
        #
        self.notebook_1.AddPage(self.nbPlaylist, _("Playlist"))
        self.notebook_1.AddPage(self.nbMenus, _("Menus"))
        self.notebook_1.AddPage(self.nbThumbnails, _("Thumbnails"))
        self.notebook_1.AddPage(self.nbDebug, _("Debug"))
        #
        # Create navigation buttons
        #
        self.btnPrevious = wx.Button(self, wx.ID_ANY, _("< Previous"))
        self.btnPrevious.Disable()
        self.btnNext = wx.Button(self, wx.ID_ANY, _("Next >"))
        self.btnFinish = wx.Button(self, wx.ID_ANY, _("Finish"))
        #
        # Add the navigation buttons
        #
        szNav = wx.BoxSizer(wx.HORIZONTAL)
        szNav.Add(self.btnPrevious, 0, wx.ADJUST_MINSIZE, 0)
        szNav.Add(self.btnNext, 0, wx.ADJUST_MINSIZE, 0)
        szNav.Add(self.btnFinish, 0, wx.ADJUST_MINSIZE, 0)
        #
        # Add elements to the screen
        #
        szFrame = wx.FlexGridSizer(2, 1, 5, 0)
        szFrame.Add(self.notebook_1, 1, wx.EXPAND, 0)
        szFrame.Add(szNav, 1, wx.ALIGN_RIGHT, 0)
        
        szFrame.Fit(self)
        szFrame.SetSizeHints(self)
        szFrame.AddGrowableRow(0)
        szFrame.AddGrowableCol(0)
        
        self.SetAutoLayout(True)
        self.SetSizer(szFrame)
        self.Layout()
        self.Centre()
        
        # global events
        self.Bind(wx.EVT_KEY_UP, self.app_OnKeyUp)
        # menu events
        wx.EVT_MENU(self, ID_MENU_FILE_EXIT, self.Close)
        wx.EVT_MENU(self, ID_MENU_VIEW_ADVANCED, self.menu_SetAdvancedView)
        wx.EVT_MENU(self, ID_MENU_HELP_ABOUT, self.menu_OnAbout)
        # button events
        wx.EVT_BUTTON(self, self.btnPrevious.GetId(), self.btnPrevious_OnClick)
        wx.EVT_BUTTON(self, self.btnNext.GetId(), self.btnNext_OnClick)
        wx.EVT_BUTTON(self, self.btnFinish.GetId(), self.btnFinish_OnClick)

    def btnPrevious_OnClick(self, evt):
        """ Event handler for clicking of '< Previous' button
        """
        currPage = self.notebook_1.GetSelection()
        if (currPage > 0):
            self.notebook_1.SetSelection(currPage - 1)
            if (not self.btnNext.IsEnabled()):
                self.btnNext.Enable()
            if (currPage - 1 == 0):
                self.btnPrevious.Disable()
        elif (self.btnPrevious.IsEnabled()):
            self.btnPrevious.Disable()

    def btnNext_OnClick(self, evt):
        """ Event handler for clicking on the 'Next >' button
        """
        currPage = self.notebook_1.GetSelection() + 1
        if (currPage < self.notebook_1.GetPageCount()):
            if (self.notebook_1.GetPage(currPage).IsEnabled()):
                self.notebook_1.SetSelection(currPage)
                if (not self.btnPrevious.IsEnabled()):
                    self.btnPrevious.Enable()
                if (currPage + 1 == self.notebook_1.GetPageCount()):
                    self.btnNext.Disable()
        elif (self.btnNext.IsEnabled()):
            self.btnNext.Disable()

    def btnFinish_OnClick(self, evt):
        """ Event handler for clicking on the 'Finish' button
        """
        # Collect todisc options from all three panels
        todisc_opts = {}
        todisc_opts.update(self.nbPlaylist.todisc_opts)
        todisc_opts.update(self.nbMenus.todisc_opts)
        todisc_opts.update(self.nbThumbnails.todisc_opts)
        # TODO: Verify required options (-files, -titles, -out)
        # Build a todisc command-line
        cmd = 'todisc '
        for opt, arg in todisc_opts.iteritems():
            # For booleans, use '-opt' (if True)
            if arg in [True, False]:
                if arg:
                    cmd += '-%s ' % opt
            # For lists, use '-opt "first" "second" ...'
            elif isinstance(arg, list):
                if len(arg) > 0:
                    cmd += '-%s ' % opt
                    for item in arg:
                        cmd += '"%s" ' % item
            # For all others, '-opt arg'
            elif arg is not None:
                cmd += '-%s %s ' % (opt, arg)
        print "Would run the following todisc command:"
        print cmd
        return

    def menu_OnAbout(self, evt):
        """Display a dialog showing information about todiscgui.
        """
        strAbout = "You are using the todisc GUI, version 0.28,\n" \
          "part of the tovid video disc authoring suite.\n\n" \
          "For more information and documentation, please\n" \
          "visit the tovid web site:\n\n" \
          "http://tovid.org/"
        dlgAbout = wx.MessageDialog(self, strAbout, "About todisc GUI", wx.OK)
        dlgAbout.ShowModal()

    def menu_SetAdvancedView(self, evt):
        if (evt.IsChecked()):
            self.notebook_1.GetPage(self.notebook_1.GetPageCount() - 1).Show()
        else:
            self.notebook_1.GetPage(self.notebook_1.GetPageCount() - 1).Hide()

    def app_OnKeyUp(self, evt):
        """Key down event handler.  Primarily used to close the app if certain keys are pressed.
        """
        key = evt.KeyCode()
        if (key >= 0 and key < 256):
            controlDown = evt.ControlDown()
            if ((controlDown) and ("Q" == chr(key))):
                self.Close()
# end of class TodiscFrame



#! /usr/bin/env python
# runtest.py

"""libtovid test script

Execute this script to run tests on the modules in libtovid. These tests
consist of executing each libtovid module standalone; it is presumed that
each module to be tested contains at least the following:

    import unittest
    if __name__ == '__main__':
        unittest.main()

which runs unit tests, or:

    import doctest
    if __name__ == '__main__':
        doctest.testmod(verbose=True)

which does an automated verification of the docstrings in each module.

NB -- You're going to get lots of output, so a >log.txt is advised.
"""

import os
import commands
from glob import glob

# Unit test modules
mod_test = glob('test/*.py')
# Library modules
# mod_libtovid = glob('*.py') ?
mod_libtovid = [\
    'cli.py',
    'deps.py',
    'layout.py',
    'media.py',
    'opts.py',
    'standard.py',
    'stats.py',
    'utils.py']
mod_author = glob('author/*.py')
mod_render = glob('render/*.py')
mod_transcode = glob('transcode/*.py')
modules = mod_test + mod_libtovid + mod_author + mod_render + mod_transcode

if __name__ == '__main__':
    # Execute each module
    for mod in modules:
        print "Testing: %s" % mod
        try:
            print commands.getoutput('python %s' % mod)
        except KeyboardInterrupt:
            print "Test interrupted."
            exit()



#! /usr/bin/env python
# flipbook.py

"""This module provides the Flipbook class.

Run this module standalone for a demonstration:

    $ python libtovid/flipbook.py

Much like a paper flipbook made of many individual drawings, the Flipbook class
has a collection of Drawing objects (from libtovid/renderers/mvg_render.py) that,
when displayed in sequence, make up an animation or video.

To use this module interactively, run 'python' from the command-line, and do:

    >>> from libtovid.flipbook import Flipbook

To create a Flipbook 10 seconds long with 'dvd' and 'ntsc' standards, do:

    >>> flipbook = Flipbook(10, 'dvd', 'ntsc', '4:3')

Then you can retrieve the canvas width and height with:

    >>> flipbook.canvas
    (640, 480)
    >>> flipbook.w
    640
    >>> flipbook.h
    480

All drawing upon a Flipbook is achieved through the use of layers
(libtovid/layer.py). To use them, import the layer module:

    >>> from libtovid import layer

You can add layers to a Flipbook on-the-fly:

    >>> flipbook.add(layer.Background('black'))
    >>> flipbook.add(layer.Text("Hello, world"))

Or, create layers separately before adding them to the Flipbook:

    >>> text = Text("Elephant talk")
    >>> flipbook.add(text)

This latter approach is useful if you plan to apply animation effects to the
Layer. For this, effects classes (libtovid/effect.py) may be applied. First,
import:

    >>> from libtovid import effect

Now, say we want the text "Elephant talk" to move across the screen, say from
(100, 100) to (500, 100) over the course of the first 90 frames. The
effect.Movement class does the trick:

    >>> text.add_effect(effect.Movement(1, 90, (100, 100), (500, 100)))

You can preview a specific frame of the Flipbook by calling render():

    >>> flipbook.render(50)

Or, you can generate all frames and create an .m2v video stream file by calling
render_video():

    >>> filename = flipbook.render_video()

See the demo below for more on what's possible with Flipbooks.
"""

__all__ = ['Flipbook']

import os
import sys
import time
import math
import cairo
import shutil
import random
from libtovid.render.animation import Keyframe
from libtovid.render.drawing import Drawing, write_ppm, write_png, interlace_drawings
from libtovid.render import layer, effect
from libtovid import standard
from libtovid.media import MediaFile, standard_media, load_media
from libtovid.transcode import encode
from libtovid import utils

class Flipbook:
    """A collection of Drawings that together comprise an animation.
    """
    def __init__(self, seconds, format, tvsys, aspect='4:3', interlaced=False):
        """Create a flipbook of the given length in seconds, at the given
        resolution.

            seconds: Length of flipbook playback in seconds
            format:  'dvd', 'vcd', 'svcd', 'dvd-vcd', or 'half-dvd'
            tvsys:   'ntsc' or 'pal'
            aspect:  '4:3' or '16:9' (the latter only for format='dvd')
                     (default: '4:3')
            interlaced: True/False. When enabled, the framerate will be
                     doubled, but only half the frames (twice the fields)
                     will be rendered.

        Once you've created the Flipbook, you can grab the dimensions
        in with the property:

            flipbook.canvas as (width, height)

        and with the two properties:

            flipbook.w (width)
            flipbook.h (height)
        """
        self.tmpdir = '/tmp'
        if 'TMPDIR' in os.environ:
            self.tmpdir = os.environ['TMPDIR']
        self.seconds = float(seconds)
        self.fps = standard.fps(tvsys)
        self.fpsratio = standard.fpsratio(tvsys)
        if (interlaced):
            self.fps *= 2
        self.interlaced = interlaced
        self.frames = int(seconds * self.fps)
        self.output_size = standard.resolution(format, tvsys)
        # TODO: We'll need aspect ratio here.. 4:3 or 16:9 anamorphic ?
        self.format = format
        self.tvsys = tvsys
        if (aspect not in ['4:3', '16:9']):
            raise ValueError, "aspect must be: '4:3' or '16:9'"
        self.aspect = aspect
        self.widescreen = False
        if (aspect == '16:9'):
            self.widescreen = True
            
        self.canvas = self._get_canvas_size()
        self.w = self.canvas[0]
        self.h = self.canvas[1]

        # Empty layers, and currently no drawings.
        self.layers = []
        self.drawings = []

        self._called_init_childs = False


    ### Child-parent handling
    def _init_childs(self):
        """Give access to all descendant layers and effects to their parents.

        In layers, you can access your parent layer (if sublayed) with:
            layer._parent_layer
        and to the top Flipbook object with:
            layer._parent_flipbook

        This function gets called just before rendering a video with
            render_video()

        It's also called in emergency if you call directly get_drawing().
        """
        for x in range(0, len(self.layers)):
            self.layers[x][0]._init_parent_flipbook(self)
            self.layers[x][0]._init_childs()

        self._called_init_childs = True
        
        
    # Utility functions

    def stof(self, seconds):
        """Return the number of frames to the specified time (in seconds)"""
        return int(self.fps * float(seconds))

    def _get_canvas_size(self):
        ow = self.output_size[0]
        oh = self.output_size[1]
        a = self.aspect.split(':')

        # Calculate the square pixels canvas size based on aspect ratio
        # and intended output size.
        x = float(a[0]) * float(oh) / float(a[1])
        # Make sure they are factor of '2'
        nw = int(math.ceil(x / 2.0) * 2)

        # We'll always change only the width, like it seems the standard,
        # since on TVs, the number of *lines* can't change (when interlacing
        # is into business)
        return (nw, oh)
        

    def add(self, layer, position=(0, 0)):
        """Add a Layer to the flipbook."""
        self.layers.append((layer, position))

    def render(self, frame=1):
        """Render the given frame."""

        print "DEPRECATED FUNCTION."
        exit()
        
        filename = "/tmp/flipbook_%s.png" % frame
        print "Rendering Flipbook frame %s to %s" % (frame, filename)
        drawing = self.get_drawing(frame)

        drawing.save_png(filename, self.output_size[0], self.output_size[1])

    def get_drawing(self, frame):
        """Get a Drawing of the given frame

        TODO: 0-based or 1-based ?
        """

        # Make sure all layers and effects has been initialized with parents.
        if (not self._called_init_childs):
            self._init_childs()
            
        drawing = Drawing(self.canvas[0], self.canvas[1])
        # Draw each layer
        for layer, position in self.layers:
            drawing.save()
            drawing.translate(position[0], position[1])
            # Apply effects and draw
            layer.draw_with_effects(drawing, frame)
            drawing.restore()
        return drawing


    def render_video(self, out_filename=None):
        """Render the flipbook to an .m2v video stream file.

            out_filename:    Filename for output.
                             If not specified, a temporary filename
                             will be given, and returned from the function.
        
        Return the filename of the output, in both cases.
        """
        ## Make sure layers and effects have been initialized with parents.
        self._init_childs()

        #
        # if self.tmpdir = /tmp, we get:
        # /tmp
        # /tmp/flipbook-120391232-work/
        # /tmp/flipbook-120391232-work/bunch-of-files..m2v, .ac3, etc.
        # /tmp/flipbook-120391232.mpg -> final output..
        #
        
        tmp_prefix = '';tmp_workdir = '';
        while 1:
            rnd = random.randint(1000000,9999999)
            tmp_prefix = "%s/flipbook-%s" % (self.tmpdir, rnd)
            tmp_workdir = "%s-work" % (tmp_prefix)

            if not os.path.exists(tmp_workdir):
                break

        # In case no filename has been specified.
        if not out_filename:
            out_filename = "%s.mpg" % tmp_workdir

        os.mkdir(tmp_workdir)

        # Encode the frames to an .m2v file
        m2v_file = "%s/video.m2v" % tmp_workdir
        ac3_file = "%s/audio.ac3" % tmp_workdir


        try:
            cairo.ImageSurface.get_data

            pngorpnm = 'ppmtoy4m -F %s -A %s -Ip -B -S 420mpeg2' % (self.fpsratio, self.aspect)
            _write_img = write_ppm  # set write function just below
        except:
            pngorpnm = 'ppmtoy4m -F %s -A %s -Ip -S 420mpeg2' % (self.fpsratio, self.aspect)
            _write_img = write_png
            
        # Cinelerra params: -b 0 -q 5 -a 2 -F 4 -I 0 -M 2 -f 8 -R 0
        # TODO: move this to encode.py
        # -M (multi-threading, num CPUs)
        #
        # -b 0 -q 5
        #    or
        # -b 7500
        pipe = os.popen("%s | mpeg2enc -q 5 -g 45 -G 45 -f 8 -a 2 -I 0 -o '%s'" % (pngorpnm,  m2v_file), 'w')
        
        frame = 1
        while frame <= self.frames:
            print "Drawing frame %s of %s" % (frame, self.frames)
            # Il faudra ici sauver DEUX drawing si on est en mode entrelacé, puis
            # les combiner en UN drawing avant de rendre en .PNG.
            # Entre les deux, on alterne les MASKS, ligne à ligne, etc..
            if (self.interlaced):
                draw1 = self.get_drawing(frame)
                frame += 1
                draw2 = self.get_drawing(frame)

                drawing = interlace_drawings(draw1, draw2)
            else:
                drawing = self.get_drawing(frame)

            # jpeg2yuv likes frames to start at 0
                _write_img(drawing, pipe, self.output_size[0], self.output_size[1], tmp_workdir)
                
            frame += 1

        print "Closing encoder. Output to %s" % out_filename
        pipe.close()

        print "Waiting for encoder to finish..."
        time.sleep(0.5)

        os.system("ls -al %s" % m2v_file)
        os.system("cp %s /tmp/go.m2v" % m2v_file)
        vidonly = load_media(m2v_file)

        # FIX the length of the load_media's bad ass..
        vidonly.length = float(self.seconds)

        outvid = MediaFile(out_filename, self.format, self.tvsys)
        outvid.aspect = self.aspect
        # Important to render 16:9 video correctly.
        outvid.widescreen = self.widescreen

        print "Running audio encoder..."
        encode.encode_audio(vidonly, ac3_file, outvid)

        print "Mplex..."
        encode.mplex_streams(m2v_file, ac3_file, outvid)

        # Clean up
        print "Cleaning up %s ..." % tmp_workdir
        shutil.rmtree(tmp_workdir)

        print "Output: %s" % out_filename
        return out_filename


        

def draw_text_demo(flipbook):
    """Draw a demonstration of Text layers with various effects."""
    assert isinstance(flipbook, Flipbook)

    last_frame = flipbook.frames
    
    # Background image
    bgd = layer.Background('black')
    flipbook.add(bgd)

    text1 = layer.Text(u"Spectrum effect demo", color=None)
    text1.add_effect(effect.Spectrum(1, last_frame))
    flipbook.add(text1, (20, 30))

    #text2 = layer.Text(u"FadeInOut effect demo")
    #text2.add_effect(effect.FadeInOut(1, last_frame, last_frame / 4))
    #flipbook.add(text2, (20, 60))

    text3 = layer.Text(u"Movement effect demo")
    text3.add_effect(effect.Movement(1, last_frame, (0, 0), (300, 50)))
    flipbook.add(text3, (20, 90))

    text4 = layer.Text(u"Whirl effect demo", color=None, align='center')
    k = [Keyframe(0, 0),
         Keyframe(30, 0),
         Keyframe(40, 15),
         Keyframe(45, -10),
         Keyframe(50, 10),
         Keyframe(55, -2),
         Keyframe(60, 0)]
    text4.add_effect(effect.Whirl(k, center=(25, 10), units='deg'))
    text4.add_effect(effect.Colorfade(1, last_frame, (1.0,0,0), (0, 1.0, 0)))
    flipbook.add(text4, (340, 350))

    text5 = layer.Text(u'Whirl effect demo2 ', color='white', align="center")
    k2 = [Keyframe(0,0),
          Keyframe(last_frame, 720)]
    text5.add_effect(effect.Whirl(k2, center=(25, 10), units='deg'))
    flipbook.add(text5, (450, 400))

    # Keyframed opacity demos, using both linear and cosine interpolation
    pulse = [Keyframe(1, 0.2),
             Keyframe(last_frame * 2/10, 1.0),
             Keyframe(last_frame * 4/10, 0.2),
             Keyframe(last_frame * 6/10, 1.0),
             Keyframe(last_frame * 8/10, 0.5),
             Keyframe(last_frame, 1.0)]
    # Text Layers
    text_linear = layer.Text("Keyframed opacity (linear)", color='lightgreen')
    text_cosine = layer.Text("Keyframed opacity (cosine)", color='lightgreen')
    # Effects
    eff_linear = effect.Fade(pulse, 'linear')
    eff_cosine = effect.Fade(pulse, 'cosine')
    # Graph of the keyframe effects
    graph_linear = layer.InterpolationGraph(pulse, method='linear')
    graph_cosine = layer.InterpolationGraph(pulse, method='cosine')
    # Add effects to the text layers
    text_linear.add_effect(eff_linear)
    text_cosine.add_effect(eff_cosine)
    # Add layers to the flipbook
    flipbook.add(text_linear, (20, 150))
    flipbook.add(graph_linear, (20, 180))
    flipbook.add(text_cosine, (340, 150))
    flipbook.add(graph_cosine, (340, 180))

# Demo
if __name__ == '__main__':
    start_time = time.time() # Benchmark

    if len(sys.argv) > 1:
        seconds = int(sys.argv[1])
    else:
        seconds = 3
    flip = Flipbook(seconds, 'dvd', 'ntsc')

    draw_text_demo(flip)

    # Video clip
    #clip = layer.VideoClip(video, (320, 240))
    #clip.rip_frames(0, 90)
    #clip.effects.append(effect.Fade(1, 90, 20))
    #flip.add(clip, (260, 200))
    
    # Render the final video
    filename = flip.render_video('/tmp/flipbook.mpg')

    print "Took %f seconds to render: %s" % (time.time() - start_time, filename)
    



#! /usr/bin/env python
# rip.py

"""Python video (audio, subtitle) ripping interface.

Currently provides:

    rip_frames:  Rip video frames from a MediaFile
    rip_video:   Rip a yuv4mpeg video stream from a MediaFile

May someday also provide:

    rip_title:    Rip video title(s) from DVD
    rip_audio:    Rip audio stream(s) from a MediaFile
    rip_subtitle: Rip subtitle(s) from a MediaFile

etc.
"""
__all__ = [\
    'rip_video',
    'rip_frames']

import os
import glob
from libtovid.cli import Command
from libtovid import log

def rip_video(source, yuvfile, target):
    """Rip video to the given yuv4mpeg file.
    
        source:  Input MediaFile
        yuvfile: File to put ripped video in
        target:  Output MediaFile
        
    """
    # TODO: Custom mplayer options, subtitles, interlacing,
    # corresp.  to $MPLAYER_OPT, $SUBTITLES, $VF_PRE/POST, $YUV4MPEG_ILACE,
    # etc.
    cmd = Command('mplayer')
    cmd.add(source.filename)
    cmd.add('-vo', 'yuv4mpeg:file=%s' % yuvfile)
    cmd.add('-nosound', '-benchmark', '-noframedrop')
    # TODO: Support subtitles. For now, use default tovid behavior.
    cmd.add('-noautosub')
    if target.scale:
        cmd.add('-vf', 'scale=%s:%s' % target.scale)
    if target.expand > target.scale:
        cmd.add('-vf-add', 'expand=%s:%s' % target.expand)
    # Run the command to rip the video
    cmd.run(background=True)


def rip_frames(media, out_dir, frames='all', size=(0, 0)):
    """Extract frame images from a MediaFile and return a list of frame image
    files.
    
        media:    MediaFile to extract images from
        out_dir:  Directory where output images should be stored; images
                  are saved in a subdirectory of out_dir named after the
                  input filename
        frames:   Which frames to rip: 'all' for all frames, 15 to rip frame
                  15; [30, 90] to rip frames 30 through 90, etc.
        size:     Resolution of frame images (default: original size), used
                  for prescaling
        
    """
    out_dir = os.path.abspath(out_dir)
    try:
        os.mkdir(out_dir)
    except:
        log.warn("Temp directory: %s already exists. Overwriting." % out_dir)
        os.system('rm -rf "%s"' % out_dir)
        os.mkdir(out_dir)

    # TODO: use tcdemux to generate a nav index, like:
    # tcdemux -f 29.970 -W -i "$FILE" > "$NAVFILE"
    
    # Use transcode to rip frames
    cmd = Command('transcode',
                  '-i', '%s' % media.filename)
    # Resize
    if size != (0, 0):
        cmd.add('-Z', '%sx%s' % size)
    # Encode selected frames
    if frames == 'all':
        start = 0
    elif isinstance(frames, int):
        cmd.add('-c', '%s-%s' % (frames, frames))
        start = frames
    elif isinstance(frames, list):
        start = frames[0]
        cmd.add('-c', '%s-%s' % (frames[0], frames[-1]))
    cmd.add('-y', 'jpg,null')
    cmd.add('-o', '%s/frame_' % out_dir)
    log.info("Creating image sequence from %s" % media.filename)
    cmd.run()

    # Return a list of ripped files
    return glob.glob('%s/frame_*.jpg' % out_dir)



#! /usr/bin/env python
# thumbmenu.py

__all__ = ['ThumbMenu']

from libtovid.render import layer, flipbook

class ThumbMenu:
    """A menu of thumbnail videos."""
    def __init__(self, target, files, titles, style):
        self.target = target
        self.files = files
        self.titles = titles
        self.style = style
        self.basename = self.target.filename
    
    def generate(self):
        print "Generating a menu with %s thumbnails" % len(self.titles)
        menu = flipbook.Flipbook(5, self.target.format, self.target.tvsys)
        width, height = self.target.scale
        thumbs = layer.ThumbGrid(self.files, self.titles,
                                 (width * 0.8, height * 0.8))
        menu.add(thumbs, (width/10, height/10))
        menu.render_video(self.basename)
        


#! /usr/bin/env python
# xml.py

"""This module is for defining XML elements and attributes, and for creating
element hierarchies.

To create a new element, use the Element class constructor, providing at least
the element name:

    >>> video = Element('video')

To see an XML representation of the Element:

    >>> print video
    <video></video>

Since this is an empty element with no attributes yet, it's pretty boring.
You can add or change attributes using the set method:

    >>> video.set(file="Brian.mpg")
    >>> print video
    <video file="Brian.mpg"></video>

To add children to an element, use the add method:

    >>> length = video.add('length', '15')
    >>> print video
    <video file="Brian.mpg">
      <length>15</length>
    </video>
    
See author.py and spumux.py for additional examples.
"""

__all__ = ['Element']

class Element (object):
    """A named XML element having optional content, attributes, and children.
    
    Attribute values may be set in the constructor, or by calling set() with a
    dictionary and/or attribute=value keywords.
    
    Use add() or add_child() to create a hierarchy of Elements.
    """
    
    def __init__(self, name, content='', attributes=None, **kwargs):
        """Create a new Element with the given attributes.
        
            name:       Name of the Element
            content:    Text content of the Element
            attributes: Dictionary of attributes and values
            kwargs:     Keyword arguments for setting specific attributes
        """
        self.name = name
        self.content = content
        self.attributes = {}
        self.children = []
        # Set attributes from those provided
        self.set(attributes, **kwargs)

    ###
    ### Public functions
    ###

    def set(self, attributes=None, **kwargs):
        """Set values for one or more attributes.
        
            attributes: Dictionary of attributes and values
            kwargs:     Keyword arguments for setting specific attributes
        
        Underscores in attribute names are converted to hyphens. If you don't
        like this behavior, please implement a workaround :-)
        """
        # Override kwargs with attributes, if provided
        if type(attributes) == dict:
            kwargs.update(attributes)
        # Set attribute values; convert underscores to hyphens
        for key, value in kwargs.iteritems():
            key = key.replace('_', '-')
            self.attributes[key] = value

    def add(self, child_name, content='', **kwargs):
        """Create and add a new child Element by providing its name, content,
        and attributes. Return the newly-added Element.
        
            child_name:  String name of child element
            content:     String content of child element
            kwargs:      Keyword arguments for setting child's attributes
        
        This is a convenience function for creating and adding children
        on-the-fly.
        """
        child = Element(child_name, content, kwargs)
        self.add_child(child)
        return child
    
    def add_child(self, element):
        """Add the given Element as a child of this Element."""
        assert isinstance(element, Element)
        self.children.append(element)

    def __str__(self):
        """Return a string containing formatted XML for the Element."""
        return self._xml().rstrip('\n')

    ###
    ### Internal functions
    ###
    
    def _open(self):
        """Return the XML opening tag for the Element."""
        attribs = ''
        for key, value in self.attributes.items():
            attribs += ' %s="%s"' % (key, value)
        return '<' + self.name + attribs + '>'

    def _close(self):
        """Return the XML closing tag for the Element."""
        return '</' + self.name + '>'

    def _xml(self, indent_level=0):
        """Return formatted XML for this Element and all descendants."""
        indent = '  ' * indent_level
        # No children; write a single line
        if len(self.children) == 0:
            text = indent + self._open() + self.content + self._close() + '\n'
        # If children, write an indented block
        else:
            text = indent + self._open() + self.content + '\n'
            for child in self.children:
                text += child._xml(indent_level + 1)
            text += indent + self._close() + '\n'
        return text



if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=True)



#! /usr/bin/env python
# textmenu.py

__all__ = ['TextMenu']

from libtovid.cli import Command, Pipe
from libtovid import deps
from libtovid import log
from libtovid import spumux

class TextMenu:
    """Simple menu with selectable text titles. For now, basically a clone
    of the classic 'makemenu' output.
    """
    def __init__(self, target, titles, style):
        deps.require('convert composite ppmtoy4m sox ffmpeg mpeg2enc')
        self.target = target
        self.titles = titles
        self.style = style
        self.basename = self.target.filename
        self.bg_canvas = self.basename + '.bg_canvas.png'
        self.fg_canvas = self.basename + '.fg_canvas.png'
        self.fg_highlight = self.basename + '.fg_highlight.png'
        self.fg_selection = self.basename + '.fg_selection.png'

    def generate(self):
        # TODO: Store intermediate images in a temp folder
        print "Creating a menu with %s titles" % len(self.titles)
        self.build_reusables()
        print "Drawing a background canvas..."
        self.draw_background_canvas()
        print "Drawing the background layer..."
        self.draw_background_layer()
        print "Drawing the highlight layer..."
        self.draw_highlight_layer()
        print "Drawing the selection layer..."
        self.draw_selection_layer()
        print "Generating the video stream..."
        self.gen_video()
        print "Generating the audio stream..."
        self.gen_audio()
        print "Multiplexing audio and video streams..."
        self.gen_mpeg()
        print "Multiplexing subtitles into .mpg..."
        self.mux_subtitles()

    def build_reusables(self):
        """Assemble some re-usable ImageMagick command snippets.
        For now, just create an ImageMagick draw command for all title text
        for placing labels and highlights.
        """
        spacing = 30    # Pixel spacing between lines
        y = 0           # Current y-coordinate
        y2 = spacing    # Next y-coordinate
        titlenum = 1        # Title number (for labeling)
        labels = ''
        buttons = ''
        for title in self.titles:
            print "Adding '%s'" % title
            # TODO: Escape special characters in title
            
            # For VCD, number the titles
            if self.target.format == 'vcd':
                labels += " text 0,%s '%s. %s' " % (y, titlenum, title)
            # Others use title string alone
            else:
                labels += " text 15,%s '%s'" % (y, title)
                buttons += " text 0,%s '>'" % y
            
            # Increment y-coordinates and title number
            y = y2
            y2 += spacing
            titlenum += 1
        self.im_labels = labels
        self.im_buttons = buttons
    
    def draw_background_canvas(self):
        """Draw background canvas."""
        # Generate default blue-black gradient background
        # TODO: Implement -background
        convert = Command('convert',
            '-size', '%sx%s' % self.target.expand,
            'gradient:blue-black',
            '-gravity', 'center',
            '-matte', self.bg_canvas)
        print "Running:", convert
        convert.run()
    
    def draw_background_layer(self):
        """Draw the background layer for the menu, including static title text.
        """
        # Draw text titles on a transparent background.
        convert = Command('convert',
            '-size', '%sx%s' % self.target.scale,
            'xc:none', '+antialias',
            '-font', self.style.font,
            '-pointsize', self.style.fontsize,
            '-fill', self.style.textcolor,
            '-stroke', 'black',
            '-strokewidth', 3,
            '-draw', 'gravity %s %s' % (self.style.align, self.im_labels),
            '-stroke', 'none',
            '-draw', 'gravity %s %s' % (self.style.align, self.im_labels),
            self.fg_canvas)
        convert.run()
        # Composite text over background
        composite = Command('composite',
            '-compose', 'Over',
            '-gravity', 'center',
            self.fg_canvas, self.bg_canvas,
            '-depth', 8, '%s.ppm' % self.basename)
        print "Running:", composite
        composite.run()
    
    def draw_highlight_layer(self):
        """Draw menu highlight layer, suitable for multiplexing."""
        # Create text layer (at safe-area size)
        convert = Command('convert',
            '-size', '%sx%s' % self.target.scale,
            'xc:none', '+antialias',
            '-font', self.style.font,
            '-weight', 'bold',
            '-pointsize', self.style.fontsize,
            '-fill', self.style.highlightcolor,
            '-draw', 'gravity %s %s' % (self.style.align, self.im_buttons),
            '-type', 'Palette', '-colors', 3,
            self.fg_highlight)
        convert.run()
        # Pseudo-composite, to expand layer to target size
        composite = Command('composite',
            '-compose', 'Src',
            '-gravity', 'center',
            self.fg_highlight, self.bg_canvas,
            '%s.hi.png' % self.basename)
        print "Running:", composite
        composite.run()
    
    def draw_selection_layer(self):
        """Draw menu selections on a transparent background."""
        # Create text layer (at safe-area size)
        convert = Command('convert',
            '-size', '%sx%s' % self.target.scale,
            'xc:none', '+antialias',
            '-font', self.style.font,
            '-weight', 'bold',
            '-pointsize', self.style.fontsize,
            '-fill', self.style.selectcolor,
            '-draw', 'gravity %s %s' % (self.style.align, self.im_buttons),
            '-type', 'Palette', '-colors', 3,
            self.fg_selection)
        convert.run()
        # Pseudo-composite, to expand layer to target size
        composite = Command('composite',
            '-compose', 'Src',
            '-gravity', 'center',
            self.fg_selection, self.bg_canvas,
            '%s.sel.png' % self.basename)
        print "Running:", composite
        composite.run()
    
    def gen_video(self):
        """Generate a video stream (mpeg1/2) from the menu background image.
        """
        # ppmtoy4m part
        ppmtoy4m = Command('ppmtoy4m', '-S', '420mpeg2')
        if self.target.tvsys == 'ntsc':
            ppmtoy4m.add('-A', '10:11', '-F', '30000:1001')
        else:
            ppmtoy4m.add('-A', '59:54', '-F', '25:1')
        # TODO: Remove hardcoded frame count
        ppmtoy4m.add('-n', 90)
        ppmtoy4m.add('-r', '%s.ppm' % self.basename)
    
        # mpeg2enc part
        mpeg2enc = Command('mpeg2enc', '-a', 2)
        # PAL/NTSC
        if self.target.tvsys == 'ntsc':
            mpeg2enc.add('-F', 4, '-n', 'n')
        else:
            mpeg2enc.add('-F', 3, '-n', 'p')
        # Use correct format flags and filename extension
        if self.target.format == 'vcd':
            self.vstream = '%s.m1v' % self.basename
            mpeg2enc.add('-f', 1)
        else:
            self.vstream = '%s.m2v' % self.basename
            if self.target.format == 'dvd':
                mpeg2enc.add('-f', 8)
            elif self.target.format == 'svcd':
                mpeg2enc.add('-f', 4)
        mpeg2enc.add('-o', self.vstream)
        pipe = Pipe(ppmtoy4m, mpeg2enc)
        print "Running:", pipe
        pipe.run()
    
    def gen_audio(self):
        """Generate an audio stream (mp2/ac3) from the given audio file
        (or generate silence instead).
        """
        if self.target.format in ['vcd', 'svcd']:
            self.astream = "%s.mp2" % self.basename
        else:
            self.astream = "%s.ac3" % self.basename
        ffmpeg = Command('ffmpeg')
        # TODO: Support including an audio stream.
        # For now, generate 4-second silence
        ffmpeg.add('-f', 's16le',
                   '-i', '/dev/zero',
                   '-t', 4)
        ffmpeg.add('-ac', 2, '-ab', 224,
                   '-ar', self.target.samprate,
                   '-acodec', self.target.acodec,
                   '-y',
                   self.astream)
        print "Running:", ffmpeg
        ffmpeg.run()
    
    def gen_mpeg(self):
        """Multiplex audio and video streams to create an mpeg.
        """
        mplex = Command('mplex', '-o', '%s.mpg' % self.basename)
        # Format flags
        if self.target.format == 'vcd':
            mplex.add('-f', 1)
        else:
            # Variable bitrate
            mplex.add('-V')
            if self.target.format == 'dvd':
                mplex.add('-f', 8)
            elif self.target.format == 'svcd':
                mplex.add('-f', 4)
        mplex.add(self.astream, self.vstream)
        print "Running:", mplex
        mplex.run()
    
    def mux_subtitles(self):
        """Multiplex the output video with highlight and selection
        subtitles, so the resulting menu can be navigated.
        """
        print "Running spumux"
        # TODO: Fix all this silly hardcoding of filenames (duplicated above)
        menu_mpg = self.basename + '.mpg'
        image = None
        select = self.basename + '.sel.png'
        highlight = self.basename + '.hi.png'
        spumux.add_subpictures(menu_mpg, image, select, highlight)



#! /usr/bin/env python
# listbox.py
# Based on grepper's ListBoxes
# under adaptation to metagui Controls

"""

Files <-> Titles (map filename to title)
Files <-> Submenu audio files (map filename to list)
Files <-> Seek values (map filename to number)

When filename in list is selected, a new control may be displayed for
setting the filename's associated attribute (title, seek, etc.)

Filenames->associated values are maintained in an Odict. Only one file's
associated-value control may be shown at a time.

When current file's associated control is updated with <Enter>, the next
filename is selected and the associated control focused.

Q. How are args assembled for associated controls? Can call get_args for
each in order, should be possible to define (at metagui top level) how they
pass options.

Speaking of top-level:

Panel("...",
    FileList('-files', "Input video files", '',
        map=Text('-titles', "Title for FILE", '')), ...
)

"""

__all__ = [
    'ListBoxes',
    'FilesTitles',
    'MappedEntry',
    'MappedFiles',
    'MappedGroups',
    'MappedNumbers',
    'MappedTitles']

import os
import sys
import time
import re
import copy
import shlex
import Tkinter as tk
import tkFileDialog
import tkMessageBox

from libtovid.metagui import *

### --------------------------------------------------------------------
### Exceptions
### --------------------------------------------------------------------

class UserError (Exception):
    """Raised when a required command-line option was not specified.

        message: Brief description of the missing option
        widget: A tkinter Widget where the option can be set
    """
    def __init__(self, message, widget=None):
        self.message = message
        self.widget = widget


### --------------------------------------------------------------------
### Helper functions
### --------------------------------------------------------------------

def blink(widget):
    """Cause a widget to "blink" by briefly changing its background color.
    """
    if widget == None:
        return
    assert isinstance(widget, tk.Widget)
    widget.config(background='#C0C0F0')
    widget.update()
    time.sleep(1)
    widget.config(background='white')


### --------------------------------------------------------------------
### Frames containing related control widgets
### --------------------------------------------------------------------

class ListBoxes (tk.Frame):
    """A frame containing a list of filenames, and controls to add or delete
    files from the list.
    """
    def __init__(self):
        self.curentry = tk.StringVar()  # currently selected mapped list entry
        self.varFiles = tk.Variable()   # List of current files
        self.varMappedList = tk.Variable()  # Current mapped list entries
        self.varUsage = tk.StringVar()  # String describing current space usage
        
        self.filesTitle = tk.StringVar()  # Title for left box ('Files' or 'Current Files')
        self.filesTitle.set('Current Files')
        
        self.mappedTitle = tk.StringVar()  # Title for the right box (mapped files)
        self.mappedTitle.set('')
        self.curindex = 0  # initialize index of currently selected file to 0
        self.puller = []  # a list of instances that the FilesTitles instance needs to sync with
        self.variable = tk.IntVar()
        self.msgSingle = tk.StringVar()
        self.msgSingle.set('')
        self.filesList = ''

    def draw(self, master):
        """Draw all the widgets in this frame."""
        tk.Frame.__init__(self, master)
        # Scrollbar to control both listboxes
        self.scrollbar = tk.Scrollbar(self, orient='vertical')
        self.scrollbar.grid(row=1, column=3, sticky='ns')
        self.scrollbar.config(command=self.scroll)
        # File list box and add/remove buttons
        self.lblFiles = tk.Label(self, textvariable=self.filesTitle)
        self.lblFiles.grid(row=0, column=0, columnspan=2, sticky='w')
        self.lstFiles = tk.Listbox(self, width=30, height=5,
                                   background='#EFEFEF',
                                   listvariable=self.varFiles,
                                   yscrollcommand=self.scrollbar.set)
        self.lstFiles.bind('<Button-1>', self.select)
        self.lstFiles.grid(row=1, column=0, columnspan=2, sticky='w')

        # Title list box and editing field
        self.lblTitles = tk.Label(self, textvariable=self.mappedTitle)
        self.lblTitles.grid(row=0, column=2, sticky='w')
        self.lstMapped = tk.Listbox(self, width=30, height=5,
                                    listvariable=self.varMappedList,
                                    yscrollcommand=self.scrollbar.set)
        self.lstMapped.grid(row=1, column=2)

    def scroll(self, *args):
        """Event handler when scrollbar is moved."""
        apply(self.lstFiles.yview, args)
        apply(self.lstMapped.yview, args)

    def select(self, event):
        """Event handler when a filename or title in the list is selected.
        Set the title box for editing and change the mouse cursor."""
        self.curindex = self.lstFiles.nearest(event.y)
        self.curentry.set(self.lstMapped.get(self.curindex))

    def makeEntry(self, event):
        """Event handler when Enter is pressed after editing a title."""
        newtitle = self.curentry.get()
        log.debug("Setting title to '%s'" % newtitle)
        self.lstMapped.delete(self.curindex)
        self.lstMapped.insert(self.curindex, newtitle)
        if self.lstFiles.get(self.curindex + 1):
            self.lstFiles.selection_clear(self.curindex)
            self.curindex = self.curindex + 1
            self.lstFiles.selection_set (self.curindex)
            self.curentry.set(self.lstMapped.get(self.curindex))

    def enableList(self):
        if self.variable.get() == 1:
            self.lstFiles.config(state='normal')
            self.syncFiles(self.filesList)
        else:
            self.lstFiles.delete(0, 'end')
            self.lstFiles.insert('end', self.msgSingle.get())
            self.lstFiles.config(state='disable')
            if self.lstMapped.get(0):
                single = self.lstMapped.get(0)
                self.lstMapped.delete(0, 'end')
                self.lstMapped.insert(0, single)
                self.curindex = 0


class FilesTitles (ListBoxes):
    """ a frame containing a listbox to show (readonly) loaded files,
    and a listbox to display and edit titles"""
    def __init__(self):
        ListBoxes.__init__(self)
        self.filesTitle.set("Files")
        self.mappedTitle.set("Titles")

    def draw(self, master):
        """Draw all the widgets in this frame."""
        ListBoxes.draw(self, master)
        self.lstMapped.bind('<B1-Motion>', self.drag)
        self.lstMapped.bind('<ButtonRelease-1>', self.drop)
        self.lstFiles.bind('<B1-Motion>', self.drag)
        self.lstFiles.bind('<ButtonRelease-1>', self.drop)
        self.lstMapped.bind('<Button-1>', self.select)
        self.lstFiles.bind('<Return>', self.makeEntry)
        self.entTitle = tk.Entry(self, width=30,
            textvariable=self.curentry)
        self.entTitle.bind('<Return>', self.makeEntry)
        self.entTitle.grid(row=2, column=2)
        self.btnAdd = tk.Button(self, text="Add...", command=self.addFiles)
        self.btnAdd.grid(row=2, column=0, sticky='ew')
        self.btnRemove = tk.Button(self, text="Remove",
                                   command=self.removeFiles)
        self.btnRemove.grid(row=2, column=1, sticky='ew')
        self.lstFiles.config(background='white')
        # Disc usage total
        self.lblUsage = tk.Label(self, textvariable=self.varUsage)
        self.lblUsage.grid(row=0, column=1, columnspan=2, sticky='e')
        self.updateUsage()

    def addFiles(self):
        """Event handler for adding files to the list box"""
        files = tkFileDialog.askopenfilenames(parent=self, title='Add files')
        for file in files:
            log.debug("Adding '%s' to the file list" % file)
            self.lstFiles.insert('end', file)
            # Add a dummy title (with pathname and extension removed)
            title = os.path.basename(file)[0:-4]
            self.lstMapped.insert('end', title)
        self.updateUsage()
        self.syncFilesList()

    def removeFiles(self):
        """Event handler for removing files from the list box"""
        selection = self.lstFiles.curselection() \
                  or self.lstMapped.curselection()
        # Using reverse order prevents reflow from messing up indexing
        if selection:
            for line in reversed(selection):
                log.debug("Removing '%s' from the file list" %\
                          self.lstFiles.get(line))
                self.lstFiles.delete(line)
                self.lstMapped.delete(line)
            self.updateUsage()
            self.syncFilesList()
            for instance_ in self.puller:
                instance_.popList(int(line))

    def addPuller(self, inst):
        """Add inst to a list of instances that sync with list of files"""
        self.puller.append(inst)

    def syncFilesList(self):
        """Sync caller instance with this instances files list"""
        for instance_ in self.puller:
            instance_.syncFiles(self.lstFiles.get(0, 'end'))

    def select(self, event):
        """Event handler when a filename or title in the list is selected.
        Set the title box for editing and change the mouse cursor."""
        self.entTitle.focus_set()
        self.curindex = self.lstFiles.nearest(event.y)
        self.curentry.set(self.lstMapped.get(self.curindex))
        self.config(cursor="double_arrow")

    def drag(self, event):
        """Event handler to move a file/title to another position in the list"""
        loc = self.lstFiles.nearest(event.y)
        if loc != self.curindex:
            file = self.lstFiles.get(self.curindex)
            mapped = self.lstMapped.get(self.curindex)
            self.lstFiles.delete(self.curindex)
            self.lstMapped.delete(self.curindex)
            self.lstFiles.insert(loc, file)
            self.lstMapped.insert(loc, mapped)
            self.curindex = loc
            self.syncFilesList()
            self.update_idletasks()

    def drop(self, event):
        """Event handler called when an item is "dropped" (mouse-release).
        Change the mouse cursor back to the default arrow.
        """
        self.config(cursor="")

    def getUsage(self):
        """Return the total size, in bytes, consumed by the current list
        of files."""
        total = 0
        for file in self.varFiles.get():
            total += os.path.getsize(file)
        return total

    def updateUsage(self):
        """Update the disc space usage label."""
        usage = self.getUsage() / (1024 * 1024)
        self.varUsage.set("%s MB used" % usage)

    def get_args(self):
        """Return todisc arguments for setting files and titles."""
        files = self.varFiles.get()
        titles = self.varMappedList.get()
        if len(files) != len(titles):
            # Should never happen, if the listboxes are properly in sync
            raise Exception, "Number of files and titles do not match"
        if len(files) == 0:
            raise UserError("File list (-files)", self.lstFiles)
        if len(titles) == 0:
            raise UserError("Title list (-titles)", self.lstMapped)
        return ['-files'] + files + ['-titles'] + titles


class MappedTitles (ListBoxes):
    """A frame containing a listbox to show (readonly) loaded files,
    and a listbox to display and edit corresponding submenu titles"""
    def __init__(self, master=None):
        ListBoxes.__init__(self)
        self.mappedTitle.set("Submenu titles")
        self.mappedfiles = []

    def draw(self, master):
        """Draw all the widgets in this frame."""
        ListBoxes.draw(self, master)
        self.entTitle = tk.Entry(self, width=30,
            textvariable=self.curentry)
        self.entTitle.bind('<Return>', self.makeEntry)
        self.lstMapped.bind('<Button-1>', self.select)
#        self.lstFiles.bind('<Double-1>', do_something)
        self.entTitle.grid(row=2, column=2)

    def syncFiles(self, filesList):
        """Sync caller instance with this instances files list"""
        self.filesList = filesList
        self.lstFiles.delete(0, 'end')
        for file in self.filesList:
            self.lstFiles.insert('end', file)
        titles = self.varMappedList.get()
        if len(titles) == 0:
            for file in filesList:
                self.lstMapped.insert('end', '')

    def select(self, event=None):
        """Event handler when a filename or title in the list is selected.
        Set the title box for editing and change the mouse cursor."""
        self.curindex = self.lstFiles.nearest(event.y)
        self.curentry.set(self.lstMapped.get(self.curindex))
        self.entTitle.focus_set()

    def popList(self, index_):
        """remove item from lstMapped list - call from FilesTitles instance"""
        self.index_ = index_
        self.lstMapped.delete(self.index_)

    def get_args(self):
        """Return todisc arguments for setting submenu titles."""
        args = []
        for title in self.varMappedList.get():
            if title:
                args.append(title)
            else:
                args.append("' '")
        return ['-submenu-titles'] + args


class MappedFiles (ListBoxes):
    """ a frame containing a listbox to show (readonly) loaded files,
    and a listbox to load files coresponding to the readonly file list"""
    def __init__(self, master=None):
        ListBoxes.__init__(self)
        self.mappedTitle.set("Submenu audio files")
        self.msgSingle.set('Use this audio file for all videos')
        self.lstFiles.insert('end', 'Use this audio file for all videos')
        self.lstFiles.config(state='disable')
        self.lstFiles.config(disabledforeground='black')

    def draw(self, master):
        """Draw all the widgets in this frame."""
        ListBoxes.draw(self, master)
        val = ''
        buttonFrame = tk.Frame(self)
        buttonFrame.grid(row=3, column=2, sticky='ew')
        self.btnAdd = tk.Button(buttonFrame, text="Add...",
                                    command=self.addFiles)
        self.btnAdd.pack(side='left', fill='x', expand=1)
        self.btnRemove = tk.Button(buttonFrame, text="Remove",
                                   command=self.removeFiles)
        self.btnRemove.pack(side='left', fill='x', expand=1)
        buttonFrame2 = tk.Frame(self)
        buttonFrame2.grid(row=3, column=0, columnspan=4, sticky='w')
        self.btnMultiple = tk.Checkbutton(buttonFrame2, text="Use multiple audio files",
                                variable=self.variable, command=self.enableList)
        self.btnMultiple.pack(side='left', fill='x', expand=1)

        self.lstFiles.bind('<Double-1>', self.addFiles)
        print self.variable.get()

    def syncFiles(self, filesList):
        """Sync list of files (lstFiles) with FilesTitles instance"""
        self.filesList = filesList
        self.lstFiles.delete(0, 'end')
        for file in self.filesList:
            self.lstFiles.insert('end', file)
        self.files = self.varMappedList.get()
        if len(self.files) == 0:
            for file in filesList:
                self.lstMapped.insert('end', '')

    def addFiles(self, event=None):
        """Event handler for adding files to the list box"""
        if not self.filesList:
            raise UserError("Load some video files first", self.lstFiles)
        else:
            self.file = tkFileDialog.askopenfilename(parent=self,
                                              title='Add files')
            this_list = list(self.varMappedList.get())
            items = len(this_list) - this_list.count('')
            if items > len(self.filesList):
                raise UserError("You have more audio files than videos!",
                                                        self.lstMapped)
            else:
                log.debug("Adding '%s' to the mapped file list" % file)
                # append files to lstMapped list for filemap
                self.lstMapped.delete(self.curindex)
                self.lstMapped.insert(self.curindex, self.file)
            # set cursor selection and index to the next file in list
            if self.lstFiles.get(self.curindex + 1):
                self.lstFiles.selection_set (self.curindex + 1)
                self.curindex = self.curindex + 1

    def removeFiles(self):
        """Event handler for removing files from the list box"""
        selection = self.lstMapped.curselection()
        # Using reverse order prevents reflow from messing up indexing
        for line in reversed(selection):
            log.debug("Removing '%s' from the file list" %\
                      self.lstFiles.get(line))
            self.lstMapped.delete(line)
            self.lstMapped.insert(line, '')

    def popList(self, index_):
        """remove item from lstMapped list - call from FilesTitles instance"""
        self.index_ = index_
        self.audiofiles = list(self.lstMapped.get(0, 'end'))
        self.numfiles = len(self.audiofiles) - self.audiofiles.count('')
        if self.numfiles != 1 or len(self.filesList) == 0:
            self.lstMapped.delete(self.index_)

    def get_args(self):
        """Return todisc arguments for setting submenu audio files."""
        self.audiofiles = list(self.lstMapped.get(0, 'end'))
        self.numfiles = len(self.audiofiles) - self.audiofiles.count('')
        if self.numfiles != 1:
            self.audio = []
            for file in self.varMappedList.get():
                if file:
                    self.audio.append(file)
                else: self.audio.append('none')
        else:
            self.audio = self.varMappedList.get()
        return ['-submenu-audio'] + self.audio


class MappedGroups (ListBoxes):
    """ a frame containing a listbox to show (readonly) loaded files, and
    a listbox to load grouped files coresponding to the readonly file list"""
    def __init__(self, master=None):
        ListBoxes.__init__(self)
        self.mappedTitle.set("These files will be grouped together")
        self.mappedfiles=[]
        self.shownIndex = 0

    def draw(self, master):
        """Draw all the widgets in this frame."""
        ListBoxes.draw(self, master)
        buttonFrame = tk.Frame(self)
        buttonFrame.grid(row=3, column=2, sticky='ew')

        self.btnAdd = tk.Button(buttonFrame, text="Add...",
                                    command=self.addFiles)
        self.btnAdd.pack(side='left', fill='x', expand=1)
        self.btnRemove = tk.Button(buttonFrame, text="Remove",
                                   command=self.removeFiles)
        self.btnRemove.pack(side='left', fill='x', expand=1)
        self.lstFiles.bind('<Double-1>', self.addFiles)
        self.lstFiles.bind('<Return>', self.addFiles)

    def select(self, event):
        """Event handler when a filename or title in the list is selected.
        A list of lists is used to allow each video to have its own group"""
        self.curindex = self.lstFiles.nearest(event.y)
        self.shownIndex = self.curindex
        # clear the tk var listMapped
        self.lstMapped.delete(0, 'end')
        # add currently selected video to group list ( ignored for cmd parsing )
        if not self.mappedfiles[self.curindex]:
            self.mappedfiles[self.curindex] = [self.filesList[self.curindex]]
        # repopulate the widget
        for file in self.mappedfiles[self.curindex]:
            if file:
                self.lstMapped.insert('end', file)

    def syncFiles(self, filesList):
        """Sync list of files (lstFiles) with FilesTitles instance"""
        self.filesList = filesList
        self.lstFiles.delete(0, 'end')
        for file in self.filesList:
            self.lstFiles.insert('end', file)
        if not self.mappedfiles:
            num_files = len(self.varFiles.get())
            self.mappedfiles = [[] for i in range( num_files)]

    def addFiles(self, event=None):
        """Event handler for adding files to the list box"""
        # clear the list for this index
        self.lstMapped.delete(0, 'end')
        if not self.filesList:
            raise UserError("Load some video files first", self.lstFiles)
        else:
            files = tkFileDialog.askopenfilenames(parent=self, title='Add files')
            for file in files:
                log.debug("Adding '%s' to the mapped file list" % file)
            # append the file to the mappedfiles list of lists
                self.mappedfiles[self.curindex].append(file)
            # update the mappedFiles widget
            for file in self.mappedfiles[self.curindex]:
                self.lstMapped.insert('end', file)
            # highlight selection to make clear which video is being grouped
            self.lstFiles.selection_set (self.curindex)

    def removeFiles(self):
        """Event handler for removing files from the list box"""
        selection = self.lstMapped.curselection()
        # Using reverse order prevents reflow from messing up indexing
        for line in reversed(selection):
            log.debug("Removing '%s' from the file list" %\
                      self.lstFiles.get(line))
            self.lstMapped.delete(0, 'end')
            self.mappedfiles[self.curindex].pop(int(line))
            file = self.mappedfiles[self.curindex]
            self.lstMapped.insert('end', *file)

    def popList(self, _index):
        """remove item from lstMapped list - call from FilesTitles instance"""
        self._index = _index
        self.mappedfiles.pop(self._index)
        if self.shownIndex == self._index:
            self.lstMapped.delete(0, 'end')

    def get_args(self):
        """Return todisc arguments for setting groups."""
        args = []
        for index, item in enumerate(self.mappedfiles):
            if item:
                args.extend(['-group', index+1] + item)

### --------------------------------------------------------------------

class MappedNumbers (ListBoxes):
    def __init__(self, master=None):
        ListBoxes.__init__(self)
        self.mappedTitle.set('Seek')
        self.lstFiles.config(disabledforeground='black')
        self.msgSingle.set('Use this seek value for all videos')
        self.lstMapped.insert(0, 2.0)

    def draw(self, master):
        ListBoxes.draw(self, master)
        self.var = tk.DoubleVar()
        self.var.set(2)
        self.curentry = tk.DoubleVar()
        self.lstMapped.config(width=6)
        NumberFrame = tk.Frame(self)
        NumberFrame.grid(row=1, column=4, sticky='ew', columnspan=2)
        self.spacer = tk.Label(NumberFrame, text="", padx=15)
        self.spacer.pack(side='left')
        self.number = tk.Scale(NumberFrame, from_=0, to=900,
                                   tickinterval=450,
                                   orient='horizontal', variable=self.var)
        self.number.pack(padx=2)
        self.entTitle = tk.Entry(NumberFrame, width=10,
            textvariable=self.var)
        self.entTitle.bind('<Return>', self.makeEntry)
        self.lstMapped.bind('<Button-1>', self.select)
        self.entTitle.pack(fill=tk.BOTH, expand=1, padx=2)
        buttonFrame = tk.Frame(self)
        buttonFrame.grid(row=2, column=0, columnspan=4, sticky='w')
        self.lstFiles.insert('end', 'Use this seek value for all videos')
        self.btnMultiple = tk.Checkbutton(buttonFrame, text="Use multiple seek values",
                                variable=self.variable, command=self.enableList)
        self.btnMultiple.pack(side='left', fill='x', expand=1)
        self.spacer1 = tk.Label(self, text="", padx=20)
        self.spacer1.grid(row=1, column=6)
        self.lstFiles.config(state='disable')
        
    def makeEntry(self, event):
        """Event handler when Enter is pressed after editing a title."""
        newtitle = self.var.get()
        log.debug("Setting title to '%s'" % newtitle)
        self.lstMapped.delete(self.curindex)
        self.lstMapped.insert(self.curindex, newtitle)
        if self.lstFiles.get(self.curindex + 1):
            self.lstFiles.selection_clear(self.curindex)
            self.curindex = self.curindex + 1
            self.lstFiles.selection_set(self.curindex)
            self.curentry.set(self.lstMapped.get(self.curindex))

    def select(self, event=None):
        """Event handler when a filename or title in the list is selected.
        Set the title box for editing and change the mouse cursor."""
        self.curindex = self.lstFiles.nearest(event.y)
        if self.lstMapped.get(self.curindex):
            self.var.set(float(self.lstMapped.get(self.curindex)))
        self.entTitle.focus_set()

    def popList(self, index_):
        """remove item from lstMapped list - call from FilesTitles instance"""
        self.index_ = index_
        self.lstMapped.delete(self.index_)
        
    def syncFiles(self, filesList):
        """Sync list of files (lstFiles) with FilesTitles instance"""
        self.filesList = filesList
        self.lstFiles.delete(0, 'end')
        for file in self.filesList:
            self.lstFiles.insert('end', file)

    def get_args(self):
        """Return todisc arguments for setting seek."""
        seeks = list(self.varMappedList.get())
        return ['-seek'] + seeks

### --------------------------------------------------------------------

class MappedEntry(ListBoxes):
    def __init__(self, master=None):
        ListBoxes.__init__(self)
        self.lstFiles.config(selectmode=tk.MULTIPLE)
        self.filesTitle.set('Select Files to chain together')
        self.mappedTitle.set('Files to be chained together')

    def draw(self, master):
        ListBoxes.draw(self, master)
        self.varChained = tk.Variable()
        self.btnAdd = tk.Button(self, text="OK", command=self.enterIndexes)
        self.btnAdd.grid(row=4, column=0, sticky='ew', columnspan=2)
        self.spacer = tk.Label(self, text="")
        self.spacer.grid(row=0, column=4, padx=2)

    def enterIndexes(self):
        self.lstMapped.delete(0, 'end')
        for i in range(0, len(self.lstFiles.get(0, 'end'))):
            self.lstMapped.insert(i, '')
        chained = ''
        self.selected = self.lstFiles.curselection()
        for index in self.selected:
            chained = chained + ' ' + str(int(index)+1)
            self.varChained.set(chained)
            self.lstMapped.delete(int(index))
            self.lstMapped.insert(int(index), self.lstFiles.get(int(index)))
        self.setHighlights()
        self.lstFiles.selection_clear(0, 'end')

    def setHighlights(self):
        mapped = self.varMappedList.get()
        for index, item in enumerate(mapped):
            if item:
                self.lstMapped.itemconfig(index, bg='#E9E7E2')

    def syncFiles(self, filesList):
        """Sync list of files (lstFiles) with FilesTitles instance"""
        self.filesList = filesList
        self.lstFiles.delete(0, 'end')
        for file in self.filesList:
            self.lstFiles.insert('end', file)

    def popList(self, index_):
        """remove item from lstMapped list - call from FilesTitles instance"""
        self.index_ = index_
        self.lstMapped.delete(self.index_)

    def get_args(self):
        """Return todisc arguments for setting chained videos."""
        chained = list(self.varChained.get())
        return ['-chain-videos'] + chained


### --------------------------------------------------------------------
### Main application window
### --------------------------------------------------------------------

class GUI (tk.Frame):
    def __init__(self, master=None):
        tk.Frame.__init__(self, master)
        self.pack()
        self.draw()

    def draw(self, master):
        """Draw all the widgets in this frame."""
        group = tk.Frame(self)
        group.pack(side='left', padx=6, pady=7)
        # Pack widgets in main frame

        self.filestitles = FilesTitles(group)
        self.filestitles.pack()
        self.spacer1 = tk.Label(group, text="")
        self.spacer1.pack()
        self.submenutitles = MappedTitles(group)
        self.submenutitles.pack()
        self.spacer2 = tk.Label(group, text="")
        self.spacer2.pack()
        self.submenuaudio = MappedFiles(group)
        self.submenuaudio.pack()
#        self.spacer3 = tk.Label(group, text="")
#        self.spacer3.pack()
        self.groupedfiles = MappedGroups(self)
        self.groupedfiles.pack(pady=7)
        self.spacer4 = tk.Label(self, text="")
        self.spacer4.pack()
        self.mappednumbers = MappedNumbers(self)
        self.mappednumbers.pack()
        self.spacer5 = tk.Label(self, text="")
        self.spacer5.pack(pady=7)
        self.mappedentry = MappedEntry(self)
        self.mappedentry.pack()
        for instance in [self.submenutitles, self.mappednumbers, self.submenuaudio,
                         self.groupedfiles, self.submenutitles, self.mappedentry]:
            self.filestitles.addPuller(instance)

        # Add the menu
        self.makeMenu()

    def makeMenu(self):
        """Create the menu for the application"""
        # create a menubar
        menubar = tk.Menu(self)
        # create a pulldown menu, and add it to the menu bar
        filemenu = tk.Menu(menubar, tearoff=False)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=root.quit)
        menubar.add_cascade(label="File", menu=filemenu)

        runmenu = tk.Menu(menubar, tearoff=False)
        runmenu.add_command(label="Run todisc now !", command=self.runCommand)
        menubar.add_cascade(label="Run", menu=runmenu)

        root.config(menu=menubar)

    def getCommand(self):
        """Return the complete todisc command."""
        cmd = Command('todisc')
        frames = [
            self.filestitles,
            self.submenuaudio,
            self.groupedfiles,
            self.mappedentry,
            self.mappednumbers,
            self.submenutitles]
        for frame in frames:
            try:
                cmd.add(*frame.get_args())
            except UserError, err:
                log.error("Missing a required option: %s" % err.message)
                raise
        cmd.add('-out')
        cmd.add('foo1234')
        return cmd

    def runCommand(self):
        """Run the todisc command."""
        try:
            cmd = self.getCommand()
        except UserError, err:
            tkMessageBox.showerror("Missing option",
                                   "Missing a required option: %s" % err.message)
            blink(err.widget)
            return
        # Show pretty-printed command
        pretty_cmd = pretty_todisc(cmd)
        log.info("Running command:")
        log.info(pretty_cmd)
        # Verify with user
        if tkMessageBox.askyesno(message="Run todisc now?"):
            root.withdraw()
            try:
                cmd.run()
            except KeyboardInterrupt:
                tkMessageBox.showerror(message="todisc was interrupted!")
            else:
                tkMessageBox.showinfo(message="todisc finished running!")
            root.deiconify()


### --------------------------------------------------------------------
### Entry point
### --------------------------------------------------------------------

root = tk.Tk()
if __name__ == '__main__':
    # Single argument: theme name 'default' or 'light'
    theme = 'light'
    if len(sys.argv) > 1:
        theme = sys.argv[1]

    root.title("todisc GUI")
    app = GUI(root)
    root.update_idletasks()
    root.mainloop()



#! /usr/bin/env python
# configs.py

import os
import sys
import re
import gettext
import wx
import locale

__all__ = ["TovidConfig"]
class TovidConfig:
    """Borg monostate class containing global configuration information"""
    __shared_state = {}

    # ==========================================================
    # Class data (will be used as if it is static)
    # ==========================================================
    # Has Config been initialized yet?
    isInitialized = False
    # Will contain fonts available in both WX and IM
    wx_IM_FontMap = {}
    # Target output directory for encoded files and finished disc
    strOutputDirectory = "/tmp"
    # XML filename used for makexml output
    strOutputXMLFile = "/tmp/untitled"
    # Is the current project a (S)VCD or DVD?
    # (TEMPORARY: Move this to a "project" class later)
    curDiscFormat = 'dvd' 
    # Reference to the GuidePanel used throughout the GUI
    panGuide = None
    # Reference to the status bar
    statusBar = None
    # The user's locale/system encoding information
    cur_lang_code, cur_encoding = locale.getdefaultlocale()
    # Hack: Error out on undefined locale
    if cur_encoding is None:
        print "Error: Could not get default locale."
        print "You should be able to fix this by setting LANG or LC_ALL:"
        print "   LANG=en_US.utf8 tovidgui, or"
        print "   LC_ALL=en_US.utf8 tovidgui"
        print "Set LANG or LC_ALL in your ~/.profile to avoid this error."
        sys.exit(1)

    def __init__(self):
        """Initialize "static" class data.
        The real initialization function, only called once."""
        self.__dict__ = self.__shared_state
        self.isInitialized = True
        self.ConfigAvailFonts()
        #self.InitLocales()
    

    def ConfigAvailFonts(self):
        """Determine fonts that are available in both wx.Python and
        ImageMagick."""
        # Find the shared fonts between ImageMagick and wx.Python
        ###########################################################
        # IM and wx store their available fonts differently, so we need a 
        # dictionary that maps the wx name for a font to its IM name. For 
        # example: "Bitstream Vera Sans Mono" in wx maps to 
        # "Bitstream-Vera-Sans-Mono" in IM. Fortunately, the difference 
        # between font names are only non-word characters. So by abbreviating
        # the names of each font, common fonts can be found in the
        # intersection of the two sets of abbreviated names. If the
        # abbreviated names are keys in a dictionary that map the abbreviated
        # name to the original name, then the shared keys between the IM and 
        # wx dictionaries can be used to build the final wx <--> IM dictionary.
        # phew!

        # Process outline:
        # 1. Find the fonts of IM and wx
        # 2. Abbreviate the names by removing spaces, underscores, and dashes
        # 3. Use dicts to map the abbreviated name (key) to the original name 
        #    (value)
        # 4. Find the common keys between both dicts (and thus the common 
        #    fonts)
        # 5. Build the wx <--> dict using the common keys

        # 1a. Find IM's fonts. IM's output is a mixture of font names and 
        #     paths. We only want the names of fonts, not the 'pretty' print 
        #     given by IM. However, since we're looking for _shared_ font 
        #     names, the 'bad' lines from IM can be left alone (as they will 
        #     be taken out later when looking for matching font names from 
        #     wx). wx is acting as our regex in a way ;) Each line that 
        #     describes a font also includes the foundry and other information
        #     we don't want. We only want the font_name, which is the 
        #     first word on each line.
        IM_lines = os.popen("convert -list type").readlines()
        IM_lines = [unicode(line, self.cur_encoding) for line in IM_lines]
        font_name_re = re.compile("^[\w-]+")   # match [a-zA-Z_-] at least once
        IM_fonts = [font_name_re.search(line).group() 
            for line in IM_lines if hasattr(font_name_re.search(line), 'group')]

        # 1b. Find wx's fonts. wx returns a list of font names as unicode 
        #     strings. This can create problems for folks with exotic fonts.
        enuWXFonts = wx.FontEnumerator()
        enuWXFonts.EnumerateFacenames()
        WX_fonts = enuWXFonts.GetFacenames()

        # 2. Remove the word separators from the font names, making 
        #    abbreviations. So:
        #    "Bitstream-Vera-Sans-Mono" becomes "BitstreamVeraSansMono" (IM)
        #    "Bitstream Vera Sans Mono" becomes "BitstreamVeraSansMono" (wx)
        abbr_name_re = re.compile("-|_| ")
        IM_abbr = [abbr_name_re.sub('', font) for font in IM_fonts]
        WX_abbr = [abbr_name_re.sub('', font) for font in WX_fonts]

        # 3. Make a dictionary map between the abbreviated name and the 
        #    original The keys are the abbreviated names, and the values are 
        #    the originals.
        #    eg:   'BitstreamVeraSansMono': 'Bitstream-Vera-Sans-Mono' (IM)
        #          'BitstreamVeraSansMono': 'Bitstream Vera Sans Mono' (wx)
        IM_dict = dict( [(IM_abbr[i], IM_fonts[i]) 
            for i in range(len(IM_fonts))] )
        WX_dict = dict( [(WX_abbr[i], WX_fonts[i]) 
            for i in range(len(WX_fonts))] )

        # 4. Find the common keys between both dicts. Make each dict's keys 
        #    a set (thereby removing any duplicate entries) and find their 
        #    intersection.
        shared_fonts = set(WX_dict.keys()).intersection(set(IM_dict.keys()))
        shared_fonts = list(shared_fonts)

        # 5. Make a dictionary that maps wx fonts to IM fonts. Using the common
        #    keys, build a dictionary that maps original wx font names to 
        #    original IM font names.
        #    eg:                   'wx_name': 'IM_name'
        #         'Bitstream Vera Sans Mono': 'Bitstream-Vera-Sans-Mono'
        WX_IM_dict = dict( [(WX_dict[font], IM_dict[font]) 
            for font in shared_fonts] )

        # Add IM's 'default' font, and return the font map
        if 'Bitstream Vera Sans' in WX_IM_dict.keys():
            WX_IM_dict['Default'] = WX_IM_dict['Bitstream Vera Sans']
        elif 'Dingbats' in WX_IM_dict.keys():
            WX_IM_dict['Default'] = WX_IM_dict['Dingbats']
        else:
            WX_IM_dict['Default'] = 'Helvetica'
        self.wx_IM_FontMap = WX_IM_dict

    def UseLanguage(self, lang):
        self.trans[ lang ].install()
    
    def InitLocales(self):
        self.trans = {}
    
        try:
            self.trans['en'] = gettext.translation('tovidgui', '.', ['en'])
            #self.trans['de'] = gettext.translation('tovidgui', '.', ['de'])
        except IOError:
            print "Couldn't initialize translations"




#! /usr/bin/env python
# effect.py

"""This module defines classes for creating and drawing effects on a series
of drawings (as in a Flipbook).

Effect classes are arranged in a (currently) simple hierarchy:

    Effect (base class)
    |-- Movement
    |   \-- Translate
    |-- Fade
    |   \-- FadeInOut
    |-- Colorfade
    |-- Spectrum
    |-- Scale
    |-- Whirl
    |-- PhotoZoom
    \-- KeyFunction

"""

__all__ = [\
    'Effect',
    'Movement',
    'Translate',
    'Fade',
    'FadeInOut'
    'Colorfade',
    'Spectrum',
    'Scale',
    'Whirl',
    'PhotoZoom',
    'KeyFunction'
    ]

from libtovid.render.drawing import Drawing
from libtovid.render.animation import Keyframe, Tween
from random import randint

class Effect:
    """A "special effect" created by keyframing a Cairo drawing command
    along the given frame interval."""
    def __init__(self, start, end):
        """Create an effect lasting from start frame to end frame."""
        self.start = start
        self.end = end
        # List of Keyframes
        self.keyframes = [Keyframe(self.start, 0), Keyframe(self.end, 0)]
        self.tween = Tween(self.keyframes)

        # Parents
        self._parent_flipbook = None
        self._parent_layer = None

    def _init_parent_flipbook(self, fb):
        self._parent_flipbook = fb

    def _init_parent_layer(self, layer):
        self._parent_layer = layer

    def pre_draw(self, drawing, frame):
        """Set up effect elements that must be applied before drawing a Layer.

            drawing: The Drawing to apply effects to
            frame:   The frame for which to render the effect

        Extend this function in derived classes.
        """
        drawing.save()

    def post_draw(self, drawing, frame):
        """Finalize effect elements after a Layer is drawn.

            drawing: The Drawing to apply effects to
            frame:   The frame for which to render the effect

        Extend this function in derived classes.
        """
        drawing.restore()


# ============================================================================
# New Effect template
# Copy and paste this code to create your own Effect
# ============================================================================
# The first line defines your effect's name. (Effect) means it inherits from
# the base Effect class, and shares some properties with it.
class MyEffect (Effect):
    """Modify this documentation string to describe what your effect does.
    """
    
    # The __init__ function is called whenever a MyEffect is created.
    # Make sure your __init__ takes start and end arguments; additional
    # arguments (such as start_val and end_val below) allow someone using
    # your effect class to customize its behavior in some way. See the
    # other effects below for examples.
    def __init__(self, start, end, start_val, end_val):
        """Create a MyEffect lasting from start to end frame.
        """
        # Be sure to call this first-thing:
        Effect.__init__(self, start, end)
        # It initializes the base Effect class with start and end frames.

        # Next, define any keyframes your effect needs to use. This
        # effect just varies something from start_val to end_val:
        self.keyframes = [\
            Keyframe(start, start_val),
            Keyframe(end, end_val)
            ]

        # Call this afterwards, to calculate the values at all frames
        self.tween = Tween(self.keyframes)

    def draw(self, drawing, frame):
        """Undocumented"""
        # Example function that can be called when drawing a layer.
        # The Layer must know that his particular effect requires the
        # calling of this function, or to add special parameters, or to
        # draw between items in the layer.
        
        # First, it's good to make sure we really have a Drawing class
        assert isinstance(drawing, Drawing)
        
        # This effect varies the stroke width across a sequence of frames.
        # Replace 'stroke_width' with your own drawing function(s)
        drawing.stroke_width(self.tween[frame])

    # Effect rendering occurs in two phases: pre_draw and post_draw.
    # These functions are already defined in the Effect base class; extend
    # them to do your effect rendering.
    def pre_draw(self, drawing, frame):
        """Do preliminary effect rendering."""
        # Always call the base class pre_draw first:
        Effect.pre_draw(self, drawing, frame)
        
        # This effect varies the stroke width across a sequence of frames.
        # Replace 'stroke_width' with your own drawing function(s)
        drawing.stroke_width(self.tween[frame])

    # Extend post_draw to do any post-rendering cleanup you need to do. Most
    # of the time, you won't need to extend this, but if you do, be sure to
    # call the base class post_draw at the end:
    def post_draw(self, drawing, frame):
        """Do post-drawing cleanup."""
        # Post-drawing cleanup here
        Effect.post_draw(self, drawing, frame)

    # That's it! Your effect is ready to use.
    # See libtovid/flipbook.py for examples on how to use effects

# ============================================================================
# End of new effect template
# ============================================================================

class Movement (Effect):
    """A movement effect, from one point to another."""
    def __init__(self, start, end, (x0, y0), (x1, y1)):
        """Move from start (x0, y0) to end (x1, y1)."""
        Effect.__init__(self, start, end)
        self.keyframes = [\
            Keyframe(start, (x0, y0)),
            Keyframe(end, (x1, y1))
            ]
        self.tween = Tween(self.keyframes)

    def pre_draw(self, drawing, frame):
        drawing.save()
        drawing.translate(*self.tween[frame])

    def post_draw(self, drawing, frame):
        drawing.restore()

class Translate (Movement):
    """Translates the layer to some relative (x,y) coordinates"""
    def __init__(self, start, end, (dx, dy)):
        Movement.__init__(self, start, end, (0,0), (dx, dy))
        

class Fade (Effect):
    """A generic fade effect, varying the opacity of a layer.
    """
    def __init__(self, keyframes, method='linear'):
        """Fade in from start, for fade_length frames; hold at full
        opacity, then fade out for fade_length frames before end.

        fade_length -- num of frames to fade-in from start, and num of
                       frames to fade-out before end. Everything in-between
                       is at full opacity.
        keyframes -- a set of Keyframe() objects, determining the fading
                     curve. Values of the Keyframe() must be floats ranging
                     from 0.0 to 1.0 (setting opacity).
        method -- linear, cosine

        """
        # A fill-opacity curve, something like:
        #         ______        100%
        #        /      \
        # start./        \.end  0%
        #
        if isinstance(keyframes, list):
            if not isinstance(keyframes[0], Keyframe):
                raise ValueError, "Must be a list of Keyframe objects"
            self.keyframes = keyframes
        else:
            raise ValueError, "List of Keyframe objects required"

        self.tween = Tween(self.keyframes, method)
        

    def pre_draw(self, drawing, frame):
        """Called before drawing on a layer"""
        drawing.push_group()

    def post_draw(self, drawing, frame):
        """Called after drawing on a layer"""
        assert isinstance(drawing, Drawing)
        drawing.pop_group_to_source()
        drawing.paint_with_alpha(self.tween[frame])

class FadeInOut(Fade):
    def __init__(self, start, end, fade_length=30):
        """Fade in from start, for fade_length frames; hold at full
        opacity, then fade out for fade_length frames before end.

        fade_length -- num of frames to fade-in from start, and num of
                       frames to fade-out before end. Everything in-between
                       is at full opacity.
        """
        # A fill-opacity curve, something like:
        #         ______        100%
        #        /      \
        # start./        \.end  0%
        #
        self.keyframes = [\
                Keyframe(start, 0.0),                  # Start fading in
                Keyframe(start + fade_length, 1.0),    # Fade-in done
                Keyframe(end - fade_length, 1.0),      # Start fading out
                Keyframe(end, 0.0)                     # Fade-out done
                ]

        self.tween = Tween(self.keyframes)

    
class Colorfade (Effect):
    """A color-slide effect between an arbitrary number of RGB colors."""
    def __init__(self, start, end, (r0, g0, b0), (r1, g1, b1)):
        """Fade between the given RGB colors."""
        Effect.__init__(self, start, end)
        self.keyframes = [\
            Keyframe(start, (r0, g0, b0)),
            Keyframe(end, (r1, g1, b1))
            ]
        self.tween = Tween(self.keyframes)

    def pre_draw(self, drawing, frame):
        """Set source color"""
        drawing.set_source(self.tween[frame])

    def post_draw(self, drawing, frame):
        pass



class Spectrum (Effect):
    """A full-spectrum color-fade effect between start and end frames."""
    def __init__(self, start, end):
        Effect.__init__(self, start, end)
        step = (end - start) / 6
        self.keyframes = [\
            Keyframe(start, (1.0, 0, 0)),
            Keyframe(start + step, (1.0, 0, 1.0)),
            Keyframe(start + step*2, (0, 0, 1.0)),
            Keyframe(start + step*3, (0, 1.0, 1.0)),
            Keyframe(start + step*4, (0, 1.0, 0)),
            Keyframe(start + step*5, (1.0, 1.0, 0)),
            Keyframe(end, (1.0, 0, 0))
            ]
        self.tween = Tween(self.keyframes)

    def pre_draw(self, drawing, frame):
        drawing.set_source(self.tween[frame])

    def post_draw(self, drawing, frame):
        pass


class Scale (Effect):
    """A Scaling effect, from one size to another."""
    def __init__(self, start, end, (w0, h0), (w1, h1)):
        Effect.__init__(self, start, end)
        self.keyframes =[\
            Keyframe(start, (w0, h0)),
            Keyframe(end, (w1, h1))
            ]
        self.tween = Tween(self.keyframes)


    def pre_draw(self, drawing, frame):
        drawing.save()
        drawing.scale(self.tween[frame])

    def post_draw(self, drawing, frame):
        drawing.restore()


class Whirl(Effect):
    """Rotates an object a number of times.
    """
    def __init__(self, keyframes, center=(0,0), method='linear', units='deg'):
        """Create a Whirl effect
        
        method -- 'linear' or 'cosine', for passing from one angle to another
                  Pass from one angle to another
        units -- 'deg' or 'rad', the unit used in the Keyframes
        """
        if units != 'deg' and units != 'rad':
            raise ValueError, "units must be 'rad' (radians) or 'deg' (degrees)"
        self.units = units
        
        if not isinstance(center, tuple):
            raise ValueError, "center must be a two-value tuple"
        self.center = center

        self.keyframes = keyframes
        self.tween = Tween(self.keyframes, method)

    def pre_draw(self, drawing, frame):
        drawing.save()
        # how to center the thing ? so you give a rotation point ?
        drawing.translate(*self.center)
        if self.units is 'deg':
            drawing.rotate_deg(self.tween[frame])
        elif self.units is 'rad':
            drawing.rotate_rad(self.tween[frame])

    def post_draw(self, drawing, frame):
        drawing.translate(- self.center[0], - self.center[1])
        drawing.restore()



class PhotoZoom(Effect):
    """Zoom in and create dynamism by moving a picture.

    Normally applies to an Image layer, but can be used with others.
    """
    def __init__(self, keyframes, subject=(0,0), direction=None, movement=50, method='linear'):
        """Create a PhotoZoom effect

        keyframes -- 0.0 for beginning of effect, and 1.0 to reach the end,
                     intermediate values show the intermediate state of the
                     effect.
        subject -- Position of the subject of interest. That is the point where
                   the zoom in/out will focus it's attention.
                   If (0,0), the focus will be chosen randomly in the 2/3 of
                   the center of the screen.
        direction -- 'in', 'out', None (random)
        movement -- 0 to 100, percentage of movement to create.
        method -- 'linear' or 'cosine'
        """
        if direction != 'in' and direction != 'out' and direction != None:
            raise ValueError, "'direction' must be 'in', 'out' or None"
        if (direction == None):
            self.direction = randint(0,1) and 'in' or 'out'
        else:
            self.direction = direction

        print "Zoom in direction: %s" % self.direction
        
        self.subject = subject

        self.movement = movement

        self.keyframes = keyframes
        self.tween = Tween(self.keyframes, method)

    def pre_draw(self, drawing, frame):
        drawing.save()

        fb = self._parent_flipbook
        # Use subject, or randomize in the center 1/3 of the image.
        if (self.subject == (0,0)):
            self.subject = (randint(int(0.33 * fb.w),
                                    int(0.66 * fb.w)),
                            randint(int(0.33 * fb.h),
                                    int(0.66 * fb.h)))

        # Max moving = 25% * pixels estimation * movement factor
        zoomfactor = 0.25 * (self.movement / 100.)

        inter = self.tween[frame]
        
        if (self.direction == 'in'):
            gozoom = 1.0 + (1.0 - inter) * zoomfactor
        else:
            gozoom = 1.0 + inter * zoomfactor


        #print "Values: zoomfactor %s inter %s gozoom %s fb %s" % (zoomfactor, inter, gozoom, fb)

        drawing.scale_centered(self.subject[0], self.subject[1],
                               gozoom, gozoom)


    def post_draw(self, drawing, frame):
        drawing.restore()




class KeyFunction (Effect):
    """A keyframed effect on an arbitrary Drawing function."""
    def __init__(self, draw_function, keyframes, method='linear'):
        """Create an effect using the given Drawing function, with values
        determined by the given list of Keyframes. For example:

            KeyFunction(Drawing.stroke_width,
                        [Keyframe(1, 1), Keyframe(30, 12)])

        This says to vary the stroke width from 1 (at frame 1) to 12 (at
        frame 30).

        The 'method' argument defines an interpolation method to use between
        keyframes, and may be either 'linear' or 'cosine'. 
        """
        # Call base constructor with start and end frames
        Effect.__init__(self, keyframes[0].frame, keyframes[-1].frame)
        # TODO: Make sure a valid function name is given
        self.draw_function = draw_function
        self.keyframes = keyframes
        # Tween keyframes using the given interpolation method
        self.tween = Tween(self.keyframes, method)

    def pre_draw(self, drawing, frame):
        drawing.save()
        self.draw_function(drawing, self.tween[frame])

    def post_draw(self, drawing, frame):
        drawing.restore()



#! /usr/bin/env python
# controls.py

import wx

import libtovid
from libtovid.gui.util import VER_GetFirstChild

__all__ = [\
    "BoldToggleButton",
    "FlexTreeCtrl",
    "HeadingText"]

class BoldToggleButton(wx.ToggleButton):
    """A wx.ToggleButton with bold font"""
    def __init__(self, parent, id, label):
        wx.ToggleButton.__init__(self, parent, id, label,
            wx.DefaultPosition, wx.Size(-1, 40))
        font = wx.Font(12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
            wx.FONTWEIGHT_BOLD)
        self.SetFont(font)


class FlexTreeCtrl(wx.TreeCtrl):
    """A more flexible tree control, with support for moving and copying
    branches."""
    
    def __init__(self, parent, id, pos = wx.DefaultPosition,
        size = wx.DefaultSize, style = wx.TR_HAS_BUTTONS,
        validator = wx.DefaultValidator, name = "FlexTreeCtrl"):
        wx.TreeCtrl.__init__(self, parent, id, pos, size, style, validator, name)
        """Initialize FlexTreeCtrl."""

    def CopyChildren(self, dest, src, recursively = True):
        """Copy all children of src to dest. Recurse if desired."""
        # Get the first child
        nextChild, cookie = VER_GetFirstChild(self, src)
        # As long as there are more children, append them to dest
        while nextChild.IsOk():
            newChild = self.AppendItem(dest,
                self.GetItemText(nextChild),
                self.GetItemImage(nextChild),
                self.GetItemImage(nextChild, wx.TreeItemIcon_Selected))
            self.SetPyData(newChild, self.GetPyData(nextChild))
            # If the child has children of its own and recursive is
            # True, call CopyChildren on this child
            if recursively and self.ItemHasChildren(nextChild):
                self.CopyChildren(newChild, nextChild, True)
            # Get the next child
            nextChild, cookie = self.GetNextChild(src, cookie)

    def MoveChildren(self, dest, src, recursively = True):
        """Recursively move all children of src to dest."""
        self.CopyChildren(dest, src, recursively)
        self.DeleteChildren(src)

    def AppendItemCopy(self, parent, src, recursively = True):
        """Append a copy of src under parent. Copy all children, and recurse if
        desired.  Returns the wx.TreeItemId of the copy, or an invalid item if
        something went wrong."""
        # Make sure parent and src are valid
        if not parent.IsOk() or not src.IsOk():
            return wx.TreeItemId()
        # Create a new copy of src under parent
        newItem = self.AppendItem(parent, self.GetItemText(src),
            self.GetItemImage(src), self.GetItemImage(src, wx.TreeItemIcon_Selected))
        self.SetPyData(newItem, self.GetPyData(src))
        # Copy children, recurse if desired
        self.CopyChildren(newItem, src, recursively)
        # Return the new copy
        return newItem

    def PrependItemCopy(self, parent, src, recursively = True):
        """Prepend a copy of src under parent. Copy all children, and recurse
        if desired.  Returns the wx.TreeItemId of the copy, or an invalid item
        if something went wrong."""
        # Make sure parent and src are valid
        if not parent.IsOk() or not src.IsOk():
            return wx.TreeItemId()
        # Create a new copy of src under parent
        newItem = self.PrependItem(parent, self.GetItemText(src),
            self.GetItemImage(src),
            self.GetItemImage(src, wx.TreeItemIcon_Selected))
        self.SetPyData(newItem, self.GetPyData(src))
        # Copy children, recurse if desired
        self.CopyChildren(newItem, src, recursively)
        # Return the new copy
        return newItem

    def InsertItemCopy(self, parent, src, prev, recursively = True):
        """Insert a copy of src under parent, after prev. Copy all children,
        and recurse if desired.  Returns the wx.TreeItemId of the copy, or an
        invalid item if something went wrong."""
        # Make sure parent, src, and prev are valid
        if not parent.IsOk() or not src.IsOk() or not prev.IsOk():
            return wx.TreeItemId()
        # Make sure prev is a child of parent
        if self.GetItemParent(prev) != parent:
            return wx.TreeItemId()
        # Create a new copy of src under parent
        newItem = self.InsertItem(parent, prev, self.GetItemText(src),
            self.GetItemImage(src),
            self.GetItemImage(src, wx.TreeItemIcon_Selected))
        self.SetPyData(newItem, self.GetPyData(src))
        # Copy children, recurse if desired
        self.CopyChildren(newItem, src, recursively)
        # Return the new copy
        return newItem

    def AppendItemMove(self, parent, src):
        """Move src, and all its descendants, under parent, at the end of
        parent's children.  If src is an ancestor of parent, do nothing.
        Returns the wx.TreeItemId of the copy, or an invalid item if something
        went wrong."""
        # If src is an ancestor of parent, return. (Cannot move
        # a branch into its own sub-branch)
        if self.ItemIsAncestorOf(src, parent):
            return wx.TreeItemId()
        # Copy src to parent and recurse
        newItem = self.AppendItemCopy(parent, src, True)
        # Delete src and all its children
        self.Delete(src)
        return newItem

    def PrependItemMove(self, parent, src):
        """Move src, and all its descendants, under parent, at the beginning of
        parent's children.  If src is an ancestor of parent, do nothing.
        Returns the wx.TreeItemId of the copy, or an invalid item if something
        went wrong."""
        # If src is an ancestor of parent, return. (Cannot move
        # a branch into its own sub-branch)
        if self.ItemIsAncestorOf(src, parent):
            return wx.TreeItemId()
        # Copy src to parent and recurse
        newItem = self.PrependItemCopy(parent, src, True)
        # Delete src and all its children
        self.Delete(src)
        return newItem

    def InsertItemMove(self, parent, src, prev):
        """Move src, and all its descendants, under parent, after parent's
        child prev.  If src is an ancestor of parent, do nothing.  Returns the
        wx.TreeItemId of the copy, or an invalid item if something went
        wrong."""
        # If src is an ancestor of parent, return. (Cannot move
        # a branch into its own sub-branch)
        if self.ItemIsAncestorOf(src, parent):
            return wx.TreeItemId()
        # Copy src to parent and recurse
        newItem = self.InsertItemCopy(parent, src, prev, True)
        # Delete src and all its children
        self.Delete(src)
        return newItem

    def ItemIsAncestorOf(self, item1, item2):
        """Return True if item1 is an ancestor of item2, False otherwise"""
        curAncestor = self.GetItemParent(item2)
        while curAncestor.IsOk():
            if curAncestor == item1:
                return True
            curAncestor = self.GetItemParent(curAncestor)

        # Root was reached and item1 was not found
        return False

    def MoveItemUp(self):
        """Move the currently-selected item up in the list.  Item stays within
        its group of siblings."""
        curItem = self.GetSelection()
        prevItem = self.GetPrevSibling(curItem)
        parentItem = self.GetItemParent(curItem)
        # If previous sibling is OK, move-insert the previous
        # sibling after the current item
        if prevItem.IsOk():
            newItem = self.InsertItemMove(parentItem, prevItem, curItem)
            if newItem.IsOk():
                self.SelectItem(curItem)

    def MoveItemDown(self):
        """Move the currently-selected item down in the list.  Item stays
        within its group of siblings."""
        curItem = self.GetSelection()

        nextItem = self.GetNextSibling(curItem)
        parentItem = self.GetItemParent(curItem)
        # If next sibling is OK, move-insert current item
        # after next sibling
        if nextItem.IsOk():
            newItem = self.InsertItemMove(parentItem, curItem, nextItem)
            if newItem.IsOk():
                self.SelectItem(newItem)

    def Delete(self, item):
        """Overridden Delete() function. Intelligently selects the previous
        item prior to deletion."""
        lastItem = item
        curParent = self.GetItemParent(item)
        prevSib = self.GetPrevSibling(lastItem)
        wx.TreeCtrl.Delete(self, lastItem)
        # Select the previous sibling, or the parent if
        # there was no previous sibling
        if prevSib.IsOk():
            self.SelectItem(prevSib)
        else:
            self.SelectItem(curParent)

    def GetReferenceList(self, topItem):
        """Returns a list of all data references in the branch descending from
        topItem."""
        # If topItem is not OK, just return
        if not topItem.IsOk():
            return
        refs = []
        # Append topItem's data
        refs.append(self.GetPyData(topItem))
        # Recursively append children
        child, cookie = VER_GetFirstChild(self, topItem)
        while child.IsOk():
            # If item has children, recurse
            if self.ItemHasChildren(child):
                grandchildren = self.GetReferenceList(child)
                refs.extend(grandchildren)
            # Otherwise, just append this item
            else:
                refs.append(self.GetPyData(child))

            # Get the next child
            child, cookie = self.GetNextChild(topItem, cookie)
        # Return the results
        return refs

    def GetItemList(self, topItem):
        """Returns a list of all Items in the branch descending from topItem."""
        # If topItem is not OK, just return
        if not topItem.IsOk():
            return
        items = []
        # Append topItem's data
        items.append(topItem)
        # Recursively append children
        child, cookie = VER_GetFirstChild(self, topItem)
        while child.IsOk():
            # If item has children, recurse
            if self.ItemHasChildren(child):
                grandchildren = self.GetItemList(child)
                items.extend(grandchildren)
            # Otherwise, just append this item
            else:
                items.append(child)

            # Get the next child
            child, cookie = self.GetNextChild(topItem, cookie)
        # Return the results
        return items

class HeadingText(wx.StaticText):
    """A large, bold static text control to use for headings"""
    def __init__(self, parent, id, label):
        wx.StaticText.__init__(self, parent, id, label)
        font = wx.Font(12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
            wx.FONTWEIGHT_BOLD)
        self.SetFont(font)

